#include "glvol.h"

/*
	OpenGL based volume renderer
	November 2003   Initial version 
	January  2004   Increased colour and histgram mappings to 65536
                   Increased histogram display to 512 wide
                   Changed the format of the colour map files, colours now RGBA
                   Added opacity and colour map editor
						 Add individual axis scaling
                   Added display list (didn't make much difference)
                   Enabled little/big endian support
                   Improved resamping performance
						 Implement non cubic inspection box
						 Animate inspection box transitions
						 Added animation of camera and colourmaps between two saved files
   March 2004      Fixed 512 pixel texture problem

	Todo
		Change autorotate to be incremental
		Add isosurface generation/display
		Add intensity scaling, how?
      Add position of cross hair cursor + world coordinates
		Add x-y plane autoscan?
	Bugs
      POsitioning of dialog windows, annoying in the center on 2 displays.
*/

#include "opacityform.c"
FD_opacityform *opacityform;
int opacityformvisible = FALSE;
#include "colourform.c"
FD_colourform *colourform;
int colourformvisible = FALSE;
#include "scaleform.c"
FD_scaleform *scaleform;
int scaleformvisible = FALSE;

OPTIONS options;
CAMERA camera;
VOLUME volume;
COLOURRAMP colourramp[11];
TEXTURE *textures = NULL;
XYZ cursor3d = {0.0,0.0,0.0};
int ntextures = 0;

RGBAI *colourmap = NULL;
int *histogram = NULL;

MARKER *markers = NULL;
int nmarkers = 0;

int currentbutton = -1;
int isshifted = FALSE;
iPOINT mousepos = {-1,-1};
int framecounter = 0;
double targetfps = 25; 
double framerate = -1;

int main(int argc,char **argv)
{
	int i,tmp;
	char camfile[128];
	XYZ origin = {0,0,0};

   /* The camera attributes */
	strcpy(camfile,"");
   camera.stereo       = NOSTEREO;
   camera.screenwidth  = 800;
   camera.screenheight = 600;
   camera.tracking     = FALSE;
   camera.autorotate   = FALSE;
   camera.autoscan     = origin;
   camera.x0 = 0;
   camera.y0 = 0;
   camera.z0 = 0;
   camera.xdim = 128;
   camera.ydim = 128;
   camera.zdim = 128;

	/* Options */
	options.debug          = FALSE;
	options.windowdump     = 0;
	options.record         = FALSE;
	options.fullscreen     = FALSE;
	options.showdecoration = TRUE;
	options.showtextures   = TRUE;
	options.showhistogram  = TRUE;
	options.showinfo       = TRUE;
   options.show3dcursor   = TRUE;
	options.showmarkers    = TRUE;
	options.xskip          = 4;
	options.yskip          = 4;
	options.zskip          = 4;
	options.animate        = FALSE;
	options.tweenframes    = 50;

   /* Default colour map and colour ramp = grey ramp */
   colourmap = malloc(MAPPINGSIZE*sizeof(RGBAI));
   for (i=0;i<MAPPINGSIZE;i++) {
      colourmap[i].r = i / (double)(MAPPINGSIZE-1);
      colourmap[i].g = i / (double)(MAPPINGSIZE-1);
      colourmap[i].b = i / (double)(MAPPINGSIZE-1);
      colourmap[i].a = i / (double)(MAPPINGSIZE-1);
   }     
   for (i=0;i<=10;i++) {
      colourramp[i].active = FALSE;
      colourramp[i].pos = i * 10;
      colourramp[i].r = i / 10.0;
      colourramp[i].g = i / 10.0;
      colourramp[i].b = i / 10.0;
      colourramp[i].a = i / 10.0;
   }
   colourramp[0].active = TRUE;
   colourramp[10].active = TRUE;
   UpdateColourMap();

   /* Initialise the volume */
   strcpy(volume.name,"untitled"); 
   strcpy(volume.comment,"");
   volume.datasize = 8;
   volume.dx = 1;
   volume.dy = 1;
   volume.dz = 1;
   volume.dim2 = CalcPower2(camera.xdim,camera.ydim,camera.zdim);
   volume.vmin =  32760;
   volume.vmax = -32768;

	/* Parse the command line arguments */
   if (argc < 2)
      GiveUsage(argv[0]);
   for (i=1;i<argc-1;i++) {
      if (strcmp(argv[i],"-h") == 0) 
         GiveUsage(argv[0]);
      if (strcmp(argv[i],"-f") == 0)
         options.fullscreen = !options.fullscreen;
      if (strcmp(argv[i],"-s") == 0)
         camera.stereo = ACTIVESTEREO;
		if (strcmp(argv[i],"-ss") == 0) {
         camera.stereo = DUALSTEREO;
			camera.screenwidth *= 2;
		}
      if (strcmp(argv[i],"-c") == 0) {
         i++;
			strcpy(camfile,argv[i]);
      }
		if (strcmp(argv[i],"-m") == 0) {
			i++;
			ReadColourMap(argv[i],0,colourmap);
			UpdateColourMap();
		}
      if (strcmp(argv[i],"-M") == 0) {
         i++;
         ReadMarkers(argv[i]);
      }
		if (strcmp(argv[i],"-a") == 0)
			camera.autorotate = TRUE;
      if (strcmp(argv[i],"-A") == 0) {
			i++;
         options.tweenframes = atoi(argv[i]);
		}
      if (strcmp(argv[i],"-r") == 0) {
         i++;
         tmp = atoi(argv[i]);
			if (tmp < 1)
				tmp = 1;
			switch (tmp) {
			case 1:
				options.xskip = 1;
				break;
			case 2:
            options.xskip = 2;
				break;
			case 3:
            options.xskip = 4;
				break;
			case 4:
            options.xskip = 8;
				break;
			}
			options.yskip = options.xskip;
			options.zskip = options.xskip;
      }
      if (strcmp(argv[i],"-vx") == 0) {
         i++;
         camera.xdim = atoi(argv[i]);
			if (camera.xdim < 16) camera.xdim = 16;
         if (camera.xdim > MAXTEXTURE) camera.xdim = MAXTEXTURE;
		   camera.xdim /= 2; camera.xdim *= 2;
		}
		if (strcmp(argv[i],"-vy") == 0) {
         i++;
         camera.ydim = atoi(argv[i]);
			if (camera.ydim < 16) camera.ydim = 16;
         if (camera.ydim > MAXTEXTURE) camera.ydim = MAXTEXTURE;
         camera.ydim /= 2; camera.ydim *= 2;
		}
      if (strcmp(argv[i],"-vz") == 0) {
         i++;
         camera.zdim = atoi(argv[i]);
         if (camera.zdim < 16) camera.zdim = 16;
         if (camera.zdim > MAXTEXTURE) camera.zdim = MAXTEXTURE;
			camera.zdim /= 2; camera.zdim *= 2;
      }
		if (strcmp(argv[i],"-d") == 0)
			options.debug = !options.debug;
   }

   /* Set things up and go */
   glutInit(&argc,argv);
   if (camera.stereo == ACTIVESTEREO)
		glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STEREO);
	else
      glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

	/* Create the window and handlers */
   glutCreateWindow("OpenGL Volume Render");
	glutReshapeWindow(camera.screenwidth,camera.screenheight);
   if (options.fullscreen)
      glutFullScreen();
   glutDisplayFunc(HandleDisplay);
	glutReshapeFunc(HandleReshape);
   glutKeyboardFunc(HandleKeyboard);
   glutSpecialFunc(HandleSpecialKeyboard);
   glutMouseFunc(HandleMouse);
	glutMotionFunc(HandleMouseMotion);
	glutPassiveMotionFunc(HandlePassiveMotion);
	glutVisibilityFunc(HandleVisibility);      
	if (camera.stereo == DUALSTEREO)
		glutSetCursor(GLUT_CURSOR_NONE); 
	glutIdleFunc(HandleIdle);
	MakeMenus();

   /* Create textures */
   textures = malloc(MAXTEXTURE*sizeof(TEXTURE));
   for (i=0;i<MAXTEXTURE;i++) {
      textures[i].pixels = NULL;
      glGenTextures(1,&(textures[i].id));
      if (textures[i].id < 1)
         fprintf(stderr,"Got bad texture id of %d for slice %d\n",(int)textures[i].id,i);
      textures[i].blank = FALSE; 
   }  
   
   /* Initialise the histogram */
   histogram = malloc(MAPPINGSIZE*sizeof(int));
   for (i=0;i<MAPPINGSIZE;i++)
      histogram[i] = 0;

   /* Get the volume */
	strcpy(volume.name,argv[argc-1]);
   ReadVol(volume.name);
   MallocTextures();
   LoadTextures(TRUE);
   CameraHome();
	if (strlen(camfile) > 1)
		ReadCameraFile(camfile,&camera);

   /* Set up the forms */
   i = 1;
   fl_initialize(&i,argv,"glVol",0,0);
   opacityform = create_form_opacityform();
   colourform  = create_form_colourform();
   scaleform   = create_form_scaleform();
   UpdateForms();

	/* Ready to go! */
   glutMainLoop();
   return(0);
}

/*
   This is the basic display callback routine
   It creates the geometry, lighting, and viewing position
*/
void HandleDisplay(void)
{
	XYZ r;
   double ratio,radians,wd2,ndfl;
   double left,right,top,bottom;

	/* Stuff for framerate calculations */
   static double tstart = -1000,tstop = 0;

   /* Set the time the first time around */
   if (tstart < 0) {
      tstart = GetRunTime();
		tstop = tstart;
	}

	camera.eyesep = camera.focallength / 30;
   ratio = camera.screenwidth / (double)camera.screenheight;
	if (camera.stereo == DUALSTEREO)
		ratio /= 2;
   radians = DTOR * camera.aperture / 2;
   wd2     = camera.near * tan(radians);
   ndfl    = camera.near / camera.focallength;
   top     =   wd2;
   bottom  = - wd2;

	/* Determine the right eye vector */
   CROSSPROD(camera.vd,camera.vu,r);
   Normalise(&r);
   r.x *= camera.eyesep / 2.0;
   r.y *= camera.eyesep / 2.0;
   r.z *= camera.eyesep / 2.0;

	/* Load textures, will only actually happen when necessary */
	LoadTextures(FALSE);

	if (camera.stereo == ACTIVESTEREO || camera.stereo == DUALSTEREO) {

      if (camera.stereo == DUALSTEREO) {
         glDrawBuffer(GL_BACK);
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      }

   	glMatrixMode(GL_PROJECTION);
   	glLoadIdentity();
      left  = - ratio * wd2 - 0.5 * camera.eyesep * ndfl;
      right =   ratio * wd2 - 0.5 * camera.eyesep * ndfl;
      glFrustum(left,right,bottom,top,camera.near,camera.far);
		if (camera.stereo == DUALSTEREO)
			glViewport(camera.screenwidth/2,0,camera.screenwidth/2,camera.screenheight);
		else
			glViewport(0,0,camera.screenwidth,camera.screenheight);
      glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK_RIGHT);
      if (camera.stereo == ACTIVESTEREO)
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glLoadIdentity();
      gluLookAt(camera.vp.x + r.x,camera.vp.y + r.y,camera.vp.z + r.z,
                camera.vp.x + r.x + camera.vd.x,
                camera.vp.y + r.y + camera.vd.y,
                camera.vp.z + r.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
      DrawAll();

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      left  = - ratio * wd2 + 0.5 * camera.eyesep * ndfl;
      right =   ratio * wd2 + 0.5 * camera.eyesep * ndfl;
      glFrustum(left,right,bottom,top,camera.near,camera.far);
      if (camera.stereo == DUALSTEREO)
         glViewport(0,0,camera.screenwidth/2,camera.screenheight);
      else
         glViewport(0,0,camera.screenwidth,camera.screenheight);
      glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK_LEFT);
      if (camera.stereo == ACTIVESTEREO)
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glLoadIdentity();
      gluLookAt(camera.vp.x - r.x,camera.vp.y - r.y,camera.vp.z - r.z,
                camera.vp.x - r.x + camera.vd.x,
                camera.vp.y - r.y + camera.vd.y,
                camera.vp.z - r.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
      DrawAll();

	} else {

     	glMatrixMode(GL_PROJECTION);
     	glLoadIdentity();
		glViewport(0,0,camera.screenwidth,camera.screenheight); 
     	left  = - ratio * wd2;
     	right =   ratio * wd2;
     	glFrustum(left,right,bottom,top,camera.near,camera.far);
     	glMatrixMode(GL_MODELVIEW);
     	glDrawBuffer(GL_BACK_LEFT);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
     	glLoadIdentity();
     	gluLookAt(camera.vp.x,camera.vp.y,camera.vp.z,
     	          camera.vp.x + camera.vd.x,
     	          camera.vp.y + camera.vd.y,
     	          camera.vp.z + camera.vd.z,
     	          camera.vu.x,camera.vu.y,camera.vu.z);
   	DrawAll();
	}

   /*
      Handle frame rate dependent features
   */
	tstop = GetRunTime();
   if (tstop - tstart > 2) {
      framerate = framecounter / (tstop - tstart);
      if (options.debug) 
         fprintf(stderr,"Framerate: %.1f frames/second\n",framerate);
      tstart = tstop;
		framecounter = 0;
	}

	/* Dump the window in tga format */
	if (options.record || options.windowdump == 1) {
		if (camera.stereo == ACTIVESTEREO)
			WindowDump("",camera.screenwidth,camera.screenheight,TRUE,12);
		else
			WindowDump("",camera.screenwidth,camera.screenheight,FALSE,12);
		options.windowdump = 0;
	}

	/* Swap buffers */
  	glutSwapBuffers();

	framecounter++;
}

/*
   Deal with plain key strokes
*/
void HandleKeyboard(unsigned char key,int x,int y)
{
	double length;

	/* Handle keyboard commands */
   switch (key) {
   case ESC: 									/* Quit */
   case 'Q':
   case 'q': 
		CleanExit();
		break;
	case 'a':                           /* Rotating */
		camera.autorotate = !camera.autorotate;
		break;
	case 'A':
		options.animate = !options.animate;
		AnimateNextFrame(TRUE);
		break;
	case 'x':                           /* Scan x axis */
		if (camera.autoscan.x != 0)
			camera.autoscan.x = 0;
		else
			camera.autoscan.x = -1;
		break;
   case 'X':
      if (camera.autoscan.x != 0)
         camera.autoscan.x = 0;
      else
         camera.autoscan.x = 1;
		break;
   case 'y':                           /* Scan y axis */
      if (camera.autoscan.y != 0)
         camera.autoscan.y = 0;
      else
         camera.autoscan.y = -1;
      break;
   case 'Y':
      if (camera.autoscan.y != 0)
         camera.autoscan.y = 0;
      else
         camera.autoscan.y = 1;
      break;
   case 'z':                           /* Scan z axis */
      if (camera.autoscan.z != 0)
         camera.autoscan.z = 0;
      else
         camera.autoscan.z = -1;
      break;
   case 'Z':
      if (camera.autoscan.z != 0)
         camera.autoscan.z = 0;
      else
         camera.autoscan.z = 1;
      break;
   case 'h':                           /* Go home     */
	case 'H':
		CameraHome();
		LoadTextures(TRUE);
      break;
   case 'i':                           /* Translate camera up */
	case 'I':
      TranslateCamera(0,1);
      break;
   case 'k':                           /* Translate camera down */
	case 'K':
      TranslateCamera(0,-1);
      break;
   case 'j':                           /* Translate camera left */
	case 'J':
      TranslateCamera(-1,0);
      break;
   case 'l':                           /* Translate camera right */
	case 'L':
      TranslateCamera(1,0);
      break;
	case 'C':                           /* Center the inspection volume */
	case 'c':
   	camera.x0 = volume.nx / 2 - camera.xdim / 2;
   	camera.y0 = volume.ny / 2 - camera.ydim / 2;
   	camera.z0 = volume.nz / 2 - camera.zdim / 2;
		LoadTextures(TRUE);
		break;
	case 'f':
	case 'F':                           /* Attach camera to the inspection volume */
		FocusCamera();
		break;
	case '<':
	case ',':
		length = VectorLength(camera.pr,camera.vp);
		camera.vp.x -= 0.01 * length * camera.vd.x;
		camera.vp.y -= 0.01 * length * camera.vd.y;
		camera.vp.z -= 0.01 * length * camera.vd.z;
		break;
   case '>':
   case '.':
      length = VectorLength(camera.pr,camera.vp);
      camera.vp.x += 0.01 * length * camera.vd.x;
      camera.vp.y += 0.01 * length * camera.vd.y;
      camera.vp.z += 0.01 * length * camera.vd.z;
      break;
   case 'w':                           /* Write the image to disk */
      options.windowdump = 1;
      break;
   case 'W':                           /* Continuous recording */
		options.record = !options.record;
		break;
	case '=':
	case '+':  // Retain volume center
		if (camera.xdim <= MAXTEXTURE/2) {
			camera.x0 -= camera.xdim/2;
			camera.xdim *= 2;
		}
		if (camera.ydim <= MAXTEXTURE/2) {
         camera.y0 -= camera.ydim/2;
			camera.ydim *= 2;
		}
		if (camera.zdim <= MAXTEXTURE/2) {
         camera.z0 -= camera.zdim/2;
         camera.zdim *= 2;
		}
		MallocTextures();
		LoadTextures(TRUE);
		break;
	case '_':
   case '-':  // Retain volume center
      if (camera.xdim > 32) {
         camera.x0 += camera.xdim/4;
			camera.xdim /= 2;
		}
		if (camera.ydim > 32) {
         camera.y0 += camera.ydim/4;
			camera.ydim /= 2;
		}
		if (camera.zdim > 32) {
         camera.z0 += camera.zdim/4;
         camera.zdim /= 2;
		}
      MallocTextures();
      LoadTextures(TRUE);
      break;
	case '1':
   	options.xskip = 1;
   	options.yskip = 1;
   	options.zskip = 1;
		LoadTextures(TRUE);
		break;
   case '2':
      options.xskip = 2;
      options.yskip = 2;
      options.zskip = 2;
		LoadTextures(TRUE);
      break;
   case '3':
      options.xskip = 4;
      options.yskip = 4;
      options.zskip = 4;
		LoadTextures(TRUE);
      break;
   case '4':
      options.xskip = 8;
      options.yskip = 8;
      options.zskip = 8;
		LoadTextures(TRUE);
      break;
	}
}

/*
   Deal with special key strokes
	Arrow keys: Shift the inspection box through the volume
*/
void HandleSpecialKeyboard(int key,int x,int y)
{
	int dx,dy,dz;

	dx = volume.nx / 100;
	if (dx < 1) dx = 1;
   if (dx > 10) dx = 10;
   dy = volume.ny / 100;
   if (dy < 1) dy = 1;
   if (dy > 10) dy = 10;
   dz = volume.nz / 100;
   if (dz < 1) dz = 1;
   if (dz > 10) dz = 10;

   switch (key) {
   case GLUT_KEY_LEFT:  
		camera.x0 -= dx;
		LoadTextures(TRUE);
		if (camera.tracking)
			FocusCamera();
		break;
   case GLUT_KEY_RIGHT: 
		camera.x0 += dx;
		LoadTextures(TRUE);
		if (camera.tracking)
			FocusCamera();
		break;
   case GLUT_KEY_UP:    
		if (glutGetModifiers() != GLUT_ACTIVE_SHIFT) {
        	camera.z0 += dz;
		} else {
			camera.y0 += dy;
		}
		LoadTextures(TRUE);
		if (camera.tracking)
			FocusCamera();
		break;
   case GLUT_KEY_DOWN:  
      if (glutGetModifiers() != GLUT_ACTIVE_SHIFT) {
        	camera.z0 -= dz;
      } else {
     		camera.y0 -= dy;
		}
		LoadTextures(TRUE);
		if (camera.tracking)
			FocusCamera();
		break;
   case GLUT_KEY_F1:
		options.showdecoration = !options.showdecoration;
		LoadTextures(TRUE);
      break;
	case GLUT_KEY_F2:
		options.showtextures = !options.showtextures;
		LoadTextures(TRUE);
		break;
   case GLUT_KEY_F3:
      options.showhistogram = !options.showhistogram;
      break;
   case GLUT_KEY_F4:
      options.showinfo = !options.showinfo;
      break;
	case GLUT_KEY_F5:
		camera.tracking = !camera.tracking;
		break;
   case GLUT_KEY_F6:
      camera.autorotate = !camera.autorotate;
      break;
   case GLUT_KEY_F7:
      options.show3dcursor = !options.show3dcursor;
      break;
   case GLUT_KEY_F8:
      options.showmarkers = !options.showmarkers;
      break;
   }
}

/*
   Rotate (ix,iy) or roll (iz) the camera about the focal point
   ix,iy,iz are flags, 0 do nothing, +- 1 rotates in opposite directions
   Correctly updating all camera attributes
*/
void RotateCamera(double ix,double iy,double iz)
{
   XYZ vp,vu,vd;
   XYZ right;
   XYZ newvp,newr;
   double radius,dd,radians;
   double dx,dy,dz;

   vu = camera.vu;
   Normalise(&vu);
   vp = camera.vp;
   vd = camera.vd;
   Normalise(&vd);
   CROSSPROD(vd,vu,right);
   Normalise(&right);
   radians = PI / 180.0;

   /* Handle the roll */
   if (iz != 0) {
      camera.vu.x += iz * right.x * radians;
      camera.vu.y += iz * right.y * radians;
      camera.vu.z += iz * right.z * radians;
      Normalise(&camera.vu);
      return;
   }

   /* Distance from the rotate point */
   dx = camera.vp.x - camera.pr.x;
   dy = camera.vp.y - camera.pr.y;
   dz = camera.vp.z - camera.pr.z;
   radius = sqrt(dx*dx + dy*dy + dz*dz);

   /* Determine the new view point */
   dd = radius * radians;
   newvp.x = vp.x + dd * ix * right.x + dd * iy * vu.x - camera.pr.x;
   newvp.y = vp.y + dd * ix * right.y + dd * iy * vu.y - camera.pr.y;
   newvp.z = vp.z + dd * ix * right.z + dd * iy * vu.z - camera.pr.z;
   Normalise(&newvp);
   camera.vp.x = camera.pr.x + radius * newvp.x;
   camera.vp.y = camera.pr.y + radius * newvp.y;
   camera.vp.z = camera.pr.z + radius * newvp.z;

   /* Determine the new right vector */
   newr.x = camera.vp.x + right.x - camera.pr.x;
   newr.y = camera.vp.y + right.y - camera.pr.y;
   newr.z = camera.vp.z + right.z - camera.pr.z;
   Normalise(&newr);
   newr.x = camera.pr.x + radius * newr.x - camera.vp.x;
   newr.y = camera.pr.y + radius * newr.y - camera.vp.y;
   newr.z = camera.pr.z + radius * newr.z - camera.vp.z;

   camera.vd.x = camera.pr.x - camera.vp.x;
   camera.vd.y = camera.pr.y - camera.vp.y;
   camera.vd.z = camera.pr.z - camera.vp.z;
   Normalise(&camera.vd);

   /* Determine the new up vector */
   CROSSPROD(newr,camera.vd,camera.vu);
   Normalise(&camera.vu);
}

/*
   Translate (pan) the camera view point
   In response to i,j,k,l keys
   Also move the camera rotate location in parallel
*/
void TranslateCamera(int ix,int iy)
{
   XYZ vp,vu,vd;
   XYZ right;
   double delta;

   vu = camera.vu;
   Normalise(&vu);
   vp = camera.vp;
   vd = camera.vd;
   Normalise(&vd);
   CROSSPROD(vd,vu,right);
   Normalise(&right);
   delta = camera.focallength / 90.0;

   camera.vp.x += iy * vu.x * delta;
   camera.vp.y += iy * vu.y * delta;
   camera.vp.z += iy * vu.z * delta;
   camera.pr.x += iy * vu.x * delta;
   camera.pr.y += iy * vu.y * delta;
   camera.pr.z += iy * vu.z * delta;

   camera.vp.x += ix * right.x * delta;
   camera.vp.y += ix * right.y * delta;
   camera.vp.z += ix * right.z * delta;
   camera.pr.x += ix * right.x * delta;
   camera.pr.y += ix * right.y * delta;
   camera.pr.z += ix * right.z * delta;
}

/*
   Handle mouse events
*/
void HandleMouse(int button,int state,int x,int y)
{
	currentbutton = -1;
   mousepos.h = x;
   mousepos.v = y;

   if (glutGetModifiers() == GLUT_ACTIVE_SHIFT)
      isshifted = TRUE;
   else
      isshifted = FALSE;

   if (state == GLUT_DOWN) {
      if (button == GLUT_LEFT_BUTTON) {
			currentbutton = GLUT_LEFT_BUTTON;
      } else if (button == GLUT_MIDDLE_BUTTON) {
			currentbutton = GLUT_MIDDLE_BUTTON;
      } else {

		}
   }
}

/*
   Handle mouse motion
	Shift key moves 3d cursor
*/
void HandleMouseMotion(int x,int y)
{
   int dx,dy;
	XYZ up,right;

   dx = x - mousepos.h;
   dy = y - mousepos.v;

	if (isshifted) {
		up = camera.vu;
		Normalise(&up);
		right = CrossProduct(camera.vd,up);
		Normalise(&right);
		cursor3d.x += dx * right.x - dy * up.x;
      cursor3d.y += dx * right.y - dy * up.y;
      cursor3d.z += dx * right.z - dy * up.z;
		cursor3d.x = MAX(0,cursor3d.x);
      cursor3d.y = MAX(0,cursor3d.y);
      cursor3d.z = MAX(0,cursor3d.z);
      cursor3d.x = MIN(volume.nx,cursor3d.x);
      cursor3d.y = MIN(volume.ny,cursor3d.y);
      cursor3d.z = MIN(volume.nz,cursor3d.z);
	} else {
   	if (currentbutton == GLUT_LEFT_BUTTON) {
   		RotateCamera(-(double)dx,(double)dy,0.0);
   	} else if (currentbutton == GLUT_MIDDLE_BUTTON) {
   	  	RotateCamera(0.0,0.0,(double)dx);
		}
	}

   mousepos.h = x;
   mousepos.v = y;
}

/*
	Collect mouse positions
*/
void HandlePassiveMotion(int x,int y)
{
   mousepos.h = x;
   mousepos.v = y;
}

/*
   What to do on an idle event
*/
void HandleIdle(void)
{
	int xnew,ynew,znew;
   static double tstart = -1;
   double tstop;
	double r,g,b,a;
	double x,y,z;
	FL_OBJECT *retobj;

	retobj = fl_check_forms();

	/* Handle scale form */
	if (scaleformvisible && retobj != NULL) {
      if (retobj == scaleform->Hide) {
         fl_hide_form(scaleform->scaleform);
         scaleformvisible = FALSE;
      }
      if (retobj == scaleform->Update) {
			x = atof(fl_get_input(scaleform->Dimx));
         y = atof(fl_get_input(scaleform->Dimy));
         z = atof(fl_get_input(scaleform->Dimz));
			if (x > 0)
				volume.dx = x;
         if (y > 0)
            volume.dy = y;
         if (z > 0)
            volume.dz = z;
         x = atof(fl_get_input(scaleform->Inspx));
         y = atof(fl_get_input(scaleform->Inspy));
         z = atof(fl_get_input(scaleform->Inspz));
         if (x < 16) x = 16;
			if (x > MAXTEXTURE) x = MAXTEXTURE;
			xnew = x/2; xnew *= 2;
         if (y < 16) y = 16;
         if (y > MAXTEXTURE) y = MAXTEXTURE;
			ynew = y/2; ynew *= 2;
         if (z < 16) z = 16;
         if (z > MAXTEXTURE) z = MAXTEXTURE;
			znew = z/2; znew *= 2;
			camera.x0 -= (xnew - camera.xdim) / 2;
         camera.y0 -= (ynew - camera.ydim) / 2;
         camera.z0 -= (znew - camera.zdim) / 2;
         camera.xdim = xnew;
			camera.ydim = ynew;
         camera.zdim = znew;
			FocusCamera();
			MallocTextures();
			LoadTextures(TRUE);
      }
	}

	/* Handle colour form events */
   if (colourformvisible && retobj != NULL) {
      if (retobj == colourform->Hide) {
         fl_hide_form(colourform->colourform);
         colourformvisible = FALSE;
      }
      if (retobj == colourform->Update) {
         colourramp[0].pos  = atof(fl_get_input(colourform->Pos0));
         colourramp[1].pos  = atof(fl_get_input(colourform->Pos10));
         colourramp[2].pos  = atof(fl_get_input(colourform->Pos20));
         colourramp[3].pos  = atof(fl_get_input(colourform->Pos30));
         colourramp[4].pos  = atof(fl_get_input(colourform->Pos40));
         colourramp[5].pos  = atof(fl_get_input(colourform->Pos50));
         colourramp[6].pos  = atof(fl_get_input(colourform->Pos60));
         colourramp[7].pos  = atof(fl_get_input(colourform->Pos70));
         colourramp[8].pos  = atof(fl_get_input(colourform->Pos80));
         colourramp[9].pos  = atof(fl_get_input(colourform->Pos90));
         colourramp[10].pos = atof(fl_get_input(colourform->Pos100));
         colourramp[0].active  = fl_get_button(colourform->Act0);
         colourramp[1].active  = fl_get_button(colourform->Act10);
         colourramp[2].active  = fl_get_button(colourform->Act20);
         colourramp[3].active  = fl_get_button(colourform->Act30);
         colourramp[4].active  = fl_get_button(colourform->Act40);
         colourramp[5].active  = fl_get_button(colourform->Act50);
         colourramp[6].active  = fl_get_button(colourform->Act60);
         colourramp[7].active  = fl_get_button(colourform->Act70);
         colourramp[8].active  = fl_get_button(colourform->Act80);
         colourramp[9].active  = fl_get_button(colourform->Act90);
         colourramp[10].active = fl_get_button(colourform->Act100);
         MapFromRamp(fl_get_button(colourform->Opacityon));
         LoadTextures(TRUE);
      }
      if (retobj == colourform->Save) {
         SaveColourRamp();
      }
      if (retobj == colourform->Read) {
         ReadColourRamp("");
      }
		if (retobj == colourform->But0  || retobj == colourform->But10 ||
          retobj == colourform->But20 || retobj == colourform->But30 ||
          retobj == colourform->But40 || retobj == colourform->But50 ||
          retobj == colourform->But60 || retobj == colourform->But70 ||
          retobj == colourform->But80 || retobj == colourform->But90 ||
          retobj == colourform->But100) {
			fl_set_button(colourform->But0,FALSE);
      	fl_set_button(colourform->But10,FALSE);
      	fl_set_button(colourform->But20,FALSE);
      	fl_set_button(colourform->But30,FALSE);
      	fl_set_button(colourform->But40,FALSE);
      	fl_set_button(colourform->But50,FALSE);
      	fl_set_button(colourform->But60,FALSE);
      	fl_set_button(colourform->But70,FALSE);
      	fl_set_button(colourform->But80,FALSE);
      	fl_set_button(colourform->But90,FALSE);
      	fl_set_button(colourform->But100,FALSE);
		}
		if (retobj == colourform->But0) {
			fl_set_slider_value(colourform->Red,colourramp[0].r);
         fl_set_slider_value(colourform->Green,colourramp[0].g);
         fl_set_slider_value(colourform->Blue,colourramp[0].b);
         fl_set_slider_value(colourform->Alpha,colourramp[0].a);
			fl_set_button(colourform->But0,TRUE);
		}
      if (retobj == colourform->But10) {
         fl_set_slider_value(colourform->Red,colourramp[1].r);
         fl_set_slider_value(colourform->Green,colourramp[1].g);
         fl_set_slider_value(colourform->Blue,colourramp[1].b);
         fl_set_slider_value(colourform->Alpha,colourramp[1].a);
         fl_set_button(colourform->But10,TRUE);
      }
      if (retobj == colourform->But20) {
         fl_set_slider_value(colourform->Red,colourramp[2].r);
         fl_set_slider_value(colourform->Green,colourramp[2].g);
         fl_set_slider_value(colourform->Blue,colourramp[2].b);
         fl_set_slider_value(colourform->Alpha,colourramp[2].a);
         fl_set_button(colourform->But20,TRUE);
      }
      if (retobj == colourform->But30) {
         fl_set_slider_value(colourform->Red,colourramp[3].r);
         fl_set_slider_value(colourform->Green,colourramp[3].g);
         fl_set_slider_value(colourform->Blue,colourramp[3].b);
         fl_set_slider_value(colourform->Alpha,colourramp[3].a);
         fl_set_button(colourform->But30,TRUE);
      }
      if (retobj == colourform->But40) {
         fl_set_slider_value(colourform->Red,colourramp[4].r);
         fl_set_slider_value(colourform->Green,colourramp[4].g);
         fl_set_slider_value(colourform->Blue,colourramp[4].b);
         fl_set_slider_value(colourform->Alpha,colourramp[4].a);
         fl_set_button(colourform->But40,TRUE);
      }
      if (retobj == colourform->But50) {
         fl_set_slider_value(colourform->Red,colourramp[5].r);
         fl_set_slider_value(colourform->Green,colourramp[5].g);
         fl_set_slider_value(colourform->Blue,colourramp[5].b);
         fl_set_slider_value(colourform->Alpha,colourramp[5].a);
         fl_set_button(colourform->But50,TRUE);
      }
      if (retobj == colourform->But60) {
         fl_set_slider_value(colourform->Red,colourramp[6].r);
         fl_set_slider_value(colourform->Green,colourramp[6].g);
         fl_set_slider_value(colourform->Blue,colourramp[6].b);
         fl_set_slider_value(colourform->Alpha,colourramp[6].a);
         fl_set_button(colourform->But60,TRUE);
      }
      if (retobj == colourform->But70) {
         fl_set_slider_value(colourform->Red,colourramp[7].r);
         fl_set_slider_value(colourform->Green,colourramp[7].g);
         fl_set_slider_value(colourform->Blue,colourramp[7].b);
         fl_set_slider_value(colourform->Alpha,colourramp[7].a);
         fl_set_button(colourform->But70,TRUE);
      }
      if (retobj == colourform->But80) {
         fl_set_slider_value(colourform->Red,colourramp[8].r);
         fl_set_slider_value(colourform->Green,colourramp[8].g);
         fl_set_slider_value(colourform->Blue,colourramp[8].b);
         fl_set_slider_value(colourform->Alpha,colourramp[8].a);
         fl_set_button(colourform->But80,TRUE);
      }
      if (retobj == colourform->But90) {
         fl_set_slider_value(colourform->Red,colourramp[9].r);
         fl_set_slider_value(colourform->Green,colourramp[9].g);
         fl_set_slider_value(colourform->Blue,colourramp[9].b);
         fl_set_slider_value(colourform->Alpha,colourramp[9].a);
         fl_set_button(colourform->But90,TRUE);
      }
      if (retobj == colourform->But100) {
         fl_set_slider_value(colourform->Red,colourramp[10].r);
         fl_set_slider_value(colourform->Green,colourramp[10].g);
         fl_set_slider_value(colourform->Blue,colourramp[10].b);
         fl_set_slider_value(colourform->Alpha,colourramp[10].a);
			fl_set_button(colourform->But100,TRUE);
      }

		if (retobj == colourform->Red || 
			 retobj == colourform->Green || 
			 retobj == colourform->Blue ||
			 retobj == colourform->Alpha) {
			r = fl_get_slider_value(colourform->Red);
         g = fl_get_slider_value(colourform->Green);
         b = fl_get_slider_value(colourform->Blue);
			a = fl_get_slider_value(colourform->Alpha);
			if (fl_get_button(colourform->But0)) {
         	fl_mapcolor(20,(int)(r*255),(int)(g*255),(int)(b*255));
         	fl_set_object_color(colourform->Col0,20,0);
         	fl_redraw_object(colourform->Col0);
				colourramp[0].r = r;
            colourramp[0].g = g;
            colourramp[0].b = b;
            colourramp[0].a = a;
			}
         if (fl_get_button(colourform->But10)) {
            fl_mapcolor(21,(int)(r*255),(int)(g*255),(int)(b*255));
            fl_set_object_color(colourform->Col10,21,0);
            fl_redraw_object(colourform->Col10);
            colourramp[1].r = r;
            colourramp[1].g = g;
            colourramp[1].b = b;
            colourramp[1].a = a;
         }
         if (fl_get_button(colourform->But20)) {
            fl_mapcolor(22,(int)(r*255),(int)(g*255),(int)(b*255));
            fl_set_object_color(colourform->Col20,22,0);
            fl_redraw_object(colourform->Col20);
            colourramp[2].r = r;
            colourramp[2].g = g;
            colourramp[2].b = b;
            colourramp[2].a = a;
         }
         if (fl_get_button(colourform->But30)) {
            fl_mapcolor(23,(int)(r*255),(int)(g*255),(int)(b*255));
            fl_set_object_color(colourform->Col30,23,0);
            fl_redraw_object(colourform->Col30);
            colourramp[3].r = r;
            colourramp[3].g = g;
            colourramp[3].b = b;
            colourramp[3].a = a;
         }
         if (fl_get_button(colourform->But40)) {
            fl_mapcolor(24,(int)(r*255),(int)(g*255),(int)(b*255));
            fl_set_object_color(colourform->Col40,24,0);
            fl_redraw_object(colourform->Col40);
            colourramp[4].r = r;
            colourramp[4].g = g;
            colourramp[4].b = b;
            colourramp[4].a = a;
         }
         if (fl_get_button(colourform->But50)) {
            fl_mapcolor(25,(int)(r*255),(int)(g*255),(int)(b*255));
            fl_set_object_color(colourform->Col50,25,0);
            fl_redraw_object(colourform->Col50);
            colourramp[5].r = r;
            colourramp[5].g = g;
            colourramp[5].b = b;
            colourramp[5].a = a;
         }
         if (fl_get_button(colourform->But60)) {
            fl_mapcolor(26,(int)(r*255),(int)(g*255),(int)(b*255));
            fl_set_object_color(colourform->Col60,26,0);
            fl_redraw_object(colourform->Col60);
            colourramp[6].r = r;
            colourramp[6].g = g;
            colourramp[6].b = b;
            colourramp[6].a = a;
         }
         if (fl_get_button(colourform->But70)) {
            fl_mapcolor(27,(int)(r*255),(int)(g*255),(int)(b*255));
            fl_set_object_color(colourform->Col70,27,0);
            fl_redraw_object(colourform->Col70);
            colourramp[7].r = r;
            colourramp[7].g = g;
            colourramp[7].b = b;
            colourramp[7].a = a;
         }
         if (fl_get_button(colourform->But80)) {
            fl_mapcolor(28,(int)(r*255),(int)(g*255),(int)(b*255));
            fl_set_object_color(colourform->Col80,28,0);
            fl_redraw_object(colourform->Col80);
            colourramp[8].r = r;
            colourramp[8].g = g;
            colourramp[8].b = b;
            colourramp[8].a = a;
         }
         if (fl_get_button(colourform->But90)) {
            fl_mapcolor(29,(int)(r*255),(int)(g*255),(int)(b*255));
            fl_set_object_color(colourform->Col90,29,0);
            fl_redraw_object(colourform->Col90);
            colourramp[9].r = r;
            colourramp[9].g = g;
            colourramp[9].b = b;
            colourramp[9].a = a;
         }
         if (fl_get_button(colourform->But100)) {
            fl_mapcolor(30,(int)(r*255),(int)(g*255),(int)(b*255));
            fl_set_object_color(colourform->Col100,30,0);
            fl_redraw_object(colourform->Col100);
            colourramp[10].r = r;
            colourramp[10].g = g;
            colourramp[10].b = b;
            colourramp[10].a = a;
         }
		}
	}

	/* Handle opacity form events */
	if (opacityformvisible && retobj != NULL) {
		if (retobj == opacityform->Hide) {
			fl_hide_form(opacityform->opacityform);
			opacityformvisible = FALSE;
		}
      if (retobj == opacityform->Update) {
			HandleOpacityForm();
			LoadTextures(TRUE);
		}
		if (retobj == opacityform->Constant) {
			fl_set_button(opacityform->Constant,TRUE);
         fl_set_button(opacityform->Linear,FALSE);
         fl_set_button(opacityform->Power,FALSE);
         fl_set_button(opacityform->Gaussian,FALSE);
		}
      if (retobj == opacityform->Linear) {
         fl_set_button(opacityform->Constant,FALSE);
         fl_set_button(opacityform->Linear,TRUE);
         fl_set_button(opacityform->Power,FALSE);
         fl_set_button(opacityform->Gaussian,FALSE);
      }
      if (retobj == opacityform->Power) {
         fl_set_button(opacityform->Constant,FALSE);
         fl_set_button(opacityform->Linear,FALSE);
         fl_set_button(opacityform->Power,TRUE);
         fl_set_button(opacityform->Gaussian,FALSE);
      }
      if (retobj == opacityform->Gaussian) {
         fl_set_button(opacityform->Constant,FALSE);
         fl_set_button(opacityform->Linear,FALSE);
         fl_set_button(opacityform->Power,FALSE);
         fl_set_button(opacityform->Gaussian,TRUE);
      }
	}

   if (tstart < 0)
      tstart = GetRunTime();
   tstop = GetRunTime();
   if (tstop - tstart > 1.0/targetfps) {
		if (options.animate) {
			AnimateNextFrame(FALSE);
			MallocTextures();
			LoadTextures(TRUE);
		}
      if (camera.autorotate)
         RotateCamera(0.1,0.0,0.0);
		if (camera.autoscan.x != 0 || camera.autoscan.y != 0 || camera.autoscan.z != 0) {
         camera.x0 += camera.autoscan.x;
         camera.y0 += camera.autoscan.y;
			camera.z0 += camera.autoscan.z;
			if (camera.x0 > volume.nx)
				camera.x0 = -camera.xdim;
         if (camera.y0 > volume.ny)
            camera.y0 = -camera.ydim;
         if (camera.z0 > volume.nz)
            camera.z0 = -camera.zdim;
         if (camera.x0 < -camera.xdim)
            camera.x0 = volume.nx;
         if (camera.y0 < -camera.ydim)
            camera.y0 = volume.ny;
         if (camera.z0 < -camera.zdim)
            camera.z0 = volume.nz;
			LoadTextures(TRUE);
      	if (camera.tracking)
         	FocusCamera();
		}
      // Nasty hack to solve Mac OS-X cursor problem
      if (camera.stereo == DUALSTEREO) {
         glutSetCursor(GLUT_CURSOR_CROSSHAIR);
         glutSetCursor(GLUT_CURSOR_NONE);
      }
      glutPostRedisplay();
      tstart = tstop;
   }

}

/* 
	Exit Cleanly
*/
void CleanExit(void)
{
	exit(0);
}

/*
   Handle a window reshape/resize
*/
void HandleReshape(int w,int h)
{
   camera.screenwidth = w;
   camera.screenheight = h;
}

/*
	Move the camera to the home position 
*/
void CameraHome(void)
{
	camera.aperture = 60;
	camera.focallength = 2*(camera.xdim*volume.dx + camera.ydim*volume.dy + camera.zdim*volume.dz)/3;
	camera.eyesep = camera.focallength / 30.0;
	camera.near = camera.focallength / 100;
	camera.far = camera.focallength * 10;

	camera.pr.x = volume.x + (camera.x0 + camera.xdim / 2) * volume.dx;
   camera.pr.y = volume.y + (camera.y0 + camera.ydim / 2) * volume.dy;
   camera.pr.z = volume.z + (camera.z0 + camera.zdim / 2) * volume.dz;
	camera.vp   = camera.pr;
	camera.vp.x -= camera.focallength;
   camera.vd.x = camera.pr.x - camera.vp.x;
   camera.vd.y = camera.pr.y - camera.vp.y;
   camera.vd.z = camera.pr.z - camera.vp.z;
   camera.vu.x = 0;
   camera.vu.y = 0;
   camera.vu.z = 1;

	Normalise(&camera.vd);
	Normalise(&camera.vu);
}
void FocusCamera(void)
{
	double dist;

   dist = VectorLength(camera.vp,camera.pr);
   camera.pr.x = volume.x + (camera.x0 + camera.xdim / 2) * volume.dx;
   camera.pr.y = volume.y + (camera.y0 + camera.ydim / 2) * volume.dy;
   camera.pr.z = volume.z + (camera.z0 + camera.zdim / 2) * volume.dz;
   camera.vp.x = camera.pr.x - dist * camera.vd.x;
   camera.vp.y = camera.pr.y - dist * camera.vd.y;
   camera.vp.z = camera.pr.z - dist * camera.vd.z;
}

int ReadVol(char *fname) 
{
	int i,j,k,c,index = 0,endian;
	FILE *fptr;
	short si;
	int li;
	double dv;

	if ((fptr = fopen(fname,"r")) == NULL) {
		fprintf(stderr,"Failed to open vol file \"%s\"\n",fname);
		CleanExit();
	}
	ReadUntil(fptr,'\n',volume.comment);
	if (fscanf(fptr,"%d %d %d",&volume.nx,&volume.ny,&volume.nz) != 3) {
		fprintf(stderr,"Failed to read dimensions\n");
		CleanExit();
	}
	if (volume.nx < 10 || volume.ny < 10 || volume.nz < 10) {
		fprintf(stderr,"Smallest dimension must be larger than 10\n");
		CleanExit();
	}
	if (options.debug)
		fprintf(stderr,"Volume resolution: %d x %d x %d\n",volume.nx,volume.ny,volume.nz);

   if (fscanf(fptr,"%lf %lf %lf",&volume.dx,&volume.dy,&volume.dz) != 3) {
      fprintf(stderr,"Failed to read cell size\n");
      CleanExit();
   }
	if (options.debug)
   	fprintf(stderr,"Voxel dimensions: %g x %g x %g\n",volume.dx,volume.dy,volume.dz);

   if (fscanf(fptr,"%lf %lf %lf",&volume.x,&volume.y,&volume.z) != 3) {
      fprintf(stderr,"Failed to read coordinate of corner\n");
      CleanExit();
   }
	if (options.debug)
		fprintf(stderr,"Volume position: %g, %g, %g\n",volume.x,volume.y,volume.z);

	if (fscanf(fptr,"%d %d",&volume.datasize,&endian) != 2) {
      fprintf(stderr,"Failed to read datasize\n");
      CleanExit();
   }
	if (volume.datasize != 8 && volume.datasize != 16 && volume.datasize != 32) {
		fprintf(stderr,"Currently only support datasize of 8, 16, or 32, not %d\n",volume.datasize);
		CleanExit();
	}
	if (options.debug) {
		fprintf(stderr,"Volume datasize: %d\n",volume.datasize);
		fprintf(stderr,"Endian: %d, I am %d\n",endian,WHATENDIAN);
	}
	SkipUntil(fptr,'\n');

	/* Malloc the data */
	VolMemory(volume.datasize,volume.nx,volume.ny,volume.nz);

	/* Read the dataset */
	volume.vmin = 1 << 30;
	volume.vmax = -volume.vmin;
	for (k=0;k<volume.nz;k++) {
		for (j=0;j<volume.ny;j++) {
			for (i=0;i<volume.nx;i++) {
				switch (volume.datasize) {
				case 8:
            	if ((c = fgetc(fptr)) == EOF) {
            	   fprintf(stderr,"Unexpected end of file at element %d,%d,%d\n",i,j,k);
            	   CleanExit();
            	}
            	volume.data1[volume.nx-i-1][volume.ny-j-1][volume.nz-k-1] = c;
					volume.vmin = MIN(volume.vmin,c);
					volume.vmax = MAX(volume.vmax,c);
					break;
				case 16:
               if (fread(&si,sizeof(short),1,fptr) != 1) {
                  fprintf(stderr,"Unexpected end of file at element %d,%d,%d\n",i,j,k);
                  CleanExit();
               }
					if (endian != WHATENDIAN)
						FIX_SHORT(si);
               volume.data2[volume.nx-i-1][volume.ny-j-1][volume.nz-k-1] = si;
               volume.vmin = MIN(volume.vmin,si);
               volume.vmax = MAX(volume.vmax,si);
               break;
            case 32:
               if (fread(&li,sizeof(int),1,fptr) != 1) {
                  fprintf(stderr,"Unexpected end of file at element %d,%d,%d\n",i,j,k);
                  CleanExit();
               }
               if (endian != WHATENDIAN)
                  FIX_LONG(li);
               volume.data3[volume.nx-i-1][volume.ny-j-1][volume.nz-k-1] = li;
               volume.vmin = MIN(volume.vmin,li);
               volume.vmax = MAX(volume.vmax,li);
               break;
				}
			}
		}
	}
	fclose(fptr);
	if (options.debug) {
		fprintf(stderr,"Volume range: %d -> %d\n",volume.vmin,volume.vmax);
	}

	/* Create the histogram */
	if (volume.datasize == 8) {
		volume.vmin = 0;
		volume.vmax = 256;
	}
	if (volume.vmax - volume.vmin < 255) 
		volume.vmax = volume.vmin + 255;
	dv = volume.vmax - volume.vmin + 1;
   for (k=0;k<volume.nz;k++) {
      for (j=0;j<volume.ny;j++) {
         for (i=0;i<volume.nx;i++) {
				switch (volume.datasize) {
				case 8:
					index = (MAPPINGSIZE/256) * volume.data1[i][j][k];
					break;
				case 16:
               index = MAPPINGSIZE * ((volume.data2[i][j][k] - volume.vmin) / dv);
					break;
            case 32:
               index = MAPPINGSIZE * ((volume.data3[i][j][k] - volume.vmin) / dv);
               break;
				}
				if (index < 0 || index >= MAPPINGSIZE) {
					fprintf(stderr,"Out of range index (%d), shouldn't happen\n",index);
					fprintf(stderr,"   %d -> %d -> %d / %g\n",volume.vmin,volume.data3[i][j][k],volume.vmax,dv);
					index = 0;
				}
				histogram[index]++;
			}
		}
	}

	/* Dream up a reasonable position for the default inspection box */
	camera.x0 = volume.nx / 2 - camera.xdim / 2;
	camera.y0 = volume.ny / 2 - camera.ydim / 2;
	camera.z0 = volume.nz / 2 - camera.zdim / 2;

	return(TRUE);
}

/*
	Create the memory for the textures
	Do this whenever the texture state changes
*/
int MallocTextures(void)
{
	int i;
	
	volume.dim2 = CalcPower2(camera.xdim,camera.ydim,camera.zdim);

	for (i=0;i<volume.dim2;i++) {
		if ((textures[i].pixels = realloc(textures[i].pixels,volume.dim2*volume.dim2*sizeof(PIXELA))) == NULL) {
			fprintf(stderr,"   Failed to malloc texture slice %d, fatal\n",i);
			exit(-1);
		}
	}

	if (options.debug)
		fprintf(stderr,"   %d textures (Each %d x %d)\n",volume.dim2,volume.dim2,volume.dim2);

	return(TRUE);
}

/*
	Load up the textures, only fill the planes actually used, see options.
	This really needs to be fast as it affects amooth mouse movement
	- Determine the view direction and hence the texture planes
	- Copy from the volume
	- Allocate textures ids
	- Bind textures
	- Record any totally blank planes
	0 == transparent, 1 (255) == opaque
*/
void LoadTextures(int force)
{
   int i,j,k,ii,jj,kk,index1,index2;
   int axis = 0;
	int nblank = 0;
	double v = 0,dv;
   XYZ cp;
	static int oldaxis = -1;

   /* Vector from center of inspection cube to camera - determines quadrant */
   cp.x = camera.vp.x - (volume.x + (camera.x0 + camera.xdim/2) * volume.dx);
   cp.y = camera.vp.y - (volume.y + (camera.y0 + camera.ydim/2) * volume.dy);
   cp.z = camera.vp.z - (volume.z + (camera.z0 + camera.zdim/2) * volume.dz);

   /* Which axis? */
   if (ABS(cp.x) >= ABS(cp.y) && ABS(cp.x) >= ABS(cp.z))
      axis = 1;
   if (ABS(cp.y) >= ABS(cp.x) && ABS(cp.y) >= ABS(cp.z))
      axis = 2;
   if (ABS(cp.z) >= ABS(cp.x) && ABS(cp.z) >= ABS(cp.y))
      axis = 3;
   if (axis == 0)
      fprintf(stderr,"Camera axis = 0, shouldn't happen\n");
	if (axis == oldaxis && !force)
		return;
	if (options.debug)
		fprintf(stderr,"New axis. Was %d, now %d\n",oldaxis,axis);
	oldaxis = axis;

	/* Set textures to black transparent */
   for (i=0;i<volume.dim2;i++) {
		memset(textures[i].pixels,0,volume.dim2*volume.dim2*4);
		textures[i].blank = TRUE;
	}

	dv = volume.vmax - volume.vmin;

	if (axis == 1) {
		for (i=0;i<camera.xdim;i+=options.xskip) {
			ii = i + camera.x0;
			if (ii < 0 || ii >= volume.nx) 
				continue;
         for (k=0;k<camera.zdim;k++) {
            kk = k + camera.z0;
            if (kk < 0 || kk >= volume.nz)
               continue;
				for (j=0;j<camera.ydim;j++) {
					jj = j + camera.y0;
					if (jj < 0 || jj >= volume.ny)
						continue;
					switch (volume.datasize) {
					case  8: v = volume.data1[ii][jj][kk]; break;
					case 16: v = volume.data2[ii][jj][kk]; break;
               case 32: v = volume.data3[ii][jj][kk]; break;
					}
					v -= volume.vmin;
					index1 = MAPPINGSIZE * v / dv;
					index2 = k*volume.dim2 + j;
					textures[i].pixels[index2] = colourmap[index1].c;
					if (colourmap[index1].c.a > 0)
						textures[i].blank = FALSE;
				}
			}
		}
	}

   if (axis == 2) {
      for (j=0;j<camera.ydim;j+=options.yskip) {
         jj = j + camera.y0;
         if (jj < 0 || jj >= volume.ny) 
            continue;
         for (k=0;k<camera.zdim;k++) {
            kk = k + camera.z0;
            if (kk < 0 || kk >= volume.nz)
               continue;
         	for (i=0;i<camera.xdim;i++) {
         	   ii = i + camera.x0;
         	   if (ii < 0 || ii >= volume.nx)
         	      continue;
               switch (volume.datasize) {
               case  8: v = volume.data1[ii][jj][kk]; break;
					case 16: v = volume.data2[ii][jj][kk]; break;
               case 32: v = volume.data3[ii][jj][kk]; break;
					}
					v -= volume.vmin;
					index1 = MAPPINGSIZE * v / dv;
					index2 = k*volume.dim2 + i;
               textures[j].pixels[index2] = colourmap[index1].c;
					if (colourmap[index1].c.a > 0)
						textures[j].blank = FALSE;
            }
         }
      }
   }

   if (axis == 3) {
      for (k=0;k<camera.zdim;k+=options.zskip) {
         kk = k + camera.z0;
         if (kk < 0 || kk >= volume.nz) 
            continue;
         for (i=0;i<camera.xdim;i++) {
            ii = i + camera.x0;
            if (ii < 0 || ii >= volume.nx)
               continue;
         	for (j=0;j<camera.ydim;j++) {
         	   jj = j + camera.y0;
         	   if (jj < 0 || jj >= volume.ny)
         	      continue;
               switch (volume.datasize) {
               case  8: v = volume.data1[ii][jj][kk]; break;
					case 16: v = volume.data2[ii][jj][kk]; break;
               case 32: v = volume.data3[ii][jj][kk]; break;
					}
					v -= volume.vmin;
					index1 = MAPPINGSIZE * v / dv;
					index2 = i*volume.dim2 + j;
               textures[k].pixels[index2] = colourmap[index1].c;
					if (colourmap[index1].c.a > 0) 
						textures[k].blank = FALSE;
            }
         }
      }
   }

	if (options.debug) {
   	for (i=0;i<volume.dim2;i++) 
   	   if (textures[i].blank)
   	      nblank++;
		fprintf(stderr,"   %d blank textures out of %d\n",nblank,volume.dim2);
	}

	/* Configure textures on card */
   for (i=0;i<volume.dim2;i++) {
   	if (textures[i].id > 0)
    	   glDeleteTextures(1,&(textures[i].id));
  		glGenTextures(1,&(textures[i].id));
		if (textures[i].id < 0)
			fprintf(stderr,"Bad texture id - probably going to be fatal, texture %d out of %d\n",
				i,volume.dim2);
      glBindTexture(GL_TEXTURE_2D,textures[i].id);
      glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
      glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
      glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
      glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
		glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,volume.dim2,volume.dim2,0,GL_RGBA,GL_UNSIGNED_BYTE,textures[i].pixels);
	}
}

/*
	Read markers
	Currently only points and lines in geom format
*/
void ReadMarkers(char *fname)
{
	char cmd[256];
	FILE *fptr;

	if ((fptr = fopen(fname,"r")) == NULL) {
		fprintf(stderr,"Failed to open marker file\n");
		return;
	}
	while (fscanf(fptr,"%s",cmd) == 1) {

		/* Points */
		if (strcmp(cmd,"p") == 0) {
			markers = realloc(markers,(nmarkers+1)*sizeof(MARKER));
			markers[nmarkers].n = 1;
			fscanf(fptr,"%lf %lf %lf",
				&(markers[nmarkers].p[0].x),&(markers[nmarkers].p[0].y),&(markers[nmarkers].p[0].z));	
         fscanf(fptr,"%lf %lf %lf",
            &(markers[nmarkers].c.r),&(markers[nmarkers].c.g),&(markers[nmarkers].c.b));
			nmarkers++;
		}

      /* Lines */
      if (strcmp(cmd,"p") == 0) {
         markers = realloc(markers,(nmarkers+1)*sizeof(MARKER));
         markers[nmarkers].n = 2;
         fscanf(fptr,"%lf %lf %lf",
            &(markers[nmarkers].p[0].x),&(markers[nmarkers].p[0].y),&(markers[nmarkers].p[0].z));
         fscanf(fptr,"%lf %lf %lf",
            &(markers[nmarkers].p[1].x),&(markers[nmarkers].p[1].y),&(markers[nmarkers].p[1].z));
         fscanf(fptr,"%lf %lf %lf",
            &(markers[nmarkers].c.r),&(markers[nmarkers].c.g),&(markers[nmarkers].c.b));
         nmarkers++;
      }

	}
	fclose(fptr);

	if (options.debug)
		fprintf(stderr,"Read %d marker objects\n",nmarkers);
}

