#include "glvol.h"

extern OPTIONS options;
extern VOLUME volume;
extern CAMERA camera;
extern RGBAI *colourmap;
extern int *histogram;

/*
   Display the program usage information
*/
void GiveUsage(char *cmd)
{
   fprintf(stderr,"Usage: %s [options] volfile\n",cmd);
   fprintf(stderr,"Options\n");
   fprintf(stderr,"          -h     this text\n");
   fprintf(stderr,"          -f     full screen\n");
   fprintf(stderr,"          -s     active stereo\n");
   fprintf(stderr,"         -ss     dual screen stereo\n");
   fprintf(stderr,"          -m s   load map file\n");
	fprintf(stderr,"          -M s   load marker file\n");
	fprintf(stderr,"          -r n   initial resolution (1,2,3,4)\n");
	fprintf(stderr,"         -vx n   initial view volume x size\n");
   fprintf(stderr,"         -vy n   initial view volume y size\n");
   fprintf(stderr,"         -vz n   initial view volume z size\n");
	fprintf(stderr,"          -c s   load camera file\n");
	fprintf(stderr,"          -a     start in auto rotate mode\n");
   fprintf(stderr,"          -A n   set tween frames for animation\n");
   fprintf(stderr,"Key Strokes\n");
   fprintf(stderr,"  arrow keys     move inspection box\n");
   fprintf(stderr,"       mouse     rotate/roll camera\n");
	fprintf(stderr,"           h     move camera to home position\n");
	fprintf(stderr,"           f     focus camera to current inspection cube\n");
	fprintf(stderr,"           c     center inspection volume\n");
	fprintf(stderr,"           w     write current window to a TGA file\n");
   fprintf(stderr,"           W     write frames to TGA file continuously\n");
   fprintf(stderr,"         i,k     translate, up/down\n");
   fprintf(stderr,"         j,l     translate left/right\n");
   fprintf(stderr," x,X,y,Y,z,Z     automatic scanning along axes\n");
   fprintf(stderr,"         <,>     move camera forward/backward\n");
   fprintf(stderr,"         +,-     change inspection box size\n");
   fprintf(stderr,"          f1     toggle decoration display\n");
   fprintf(stderr,"          f2     toggle texture display\n");
   fprintf(stderr,"          f3     toggle histogram display\n");
   fprintf(stderr,"          f4     toggle text information\n");
   fprintf(stderr,"          f5     toggle camera tracking\n");
   fprintf(stderr,"          f6     toggle autorotate\n");
	fprintf(stderr,"          f7     toggle 3d cursor\n");
   fprintf(stderr,"     1,2,3,4     set supsampling level to 1,2,4,8\n");
   fprintf(stderr,"           q     quit\n");
   exit(-1);
}

/*
   This is where global OpenGL/GLUT settings are made,
   that is, things that will not change in time
*/
void CreateEnvironment(void)
{
	static int first = TRUE;
	GLint num[10];

   glEnable(GL_DEPTH_TEST);
   glEnable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

   glDisable(GL_DITHER);
   glDisable(GL_LINE_SMOOTH);
   glDisable(GL_POINT_SMOOTH);
   glDisable(GL_POLYGON_SMOOTH);

   glLineWidth(1.0);
   glPointSize(1.0);

   glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
   glShadeModel(GL_SMOOTH);
   glFrontFace(GL_CW);
   glDisable(GL_CULL_FACE);

   glClearColor(0.0,0.0,0.0,0.0);
   glPixelStorei(GL_UNPACK_ALIGNMENT,1);

   glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
   glEnable(GL_COLOR_MATERIAL);
   glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_FALSE);
   glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);

   glDisable(GL_LIGHTING);

	if (first) {
   	// Sort out multisampling
   	glGetIntegerv(GL_SAMPLE_BUFFERS,&(num[0]));
   	glGetIntegerv(GL_SAMPLES,&(num[1]));
   	if (options.debug) {
   	   fprintf(stderr,"Sample buffers: %d, samples: %d\n",(int)num[0],(int)num[1]);
   	   if (num[0] > 0 && num[1] > 1)
   	      fprintf(stderr,"Multisampling enabled\n");
   	   else
   	      fprintf(stderr,"Multisampling not enabled\n");
   	}
   	if (num[0] > 0 && num[1] > 1)
   	   glEnable(GL_MULTISAMPLE);
	}

	first = FALSE;
}

/*
   How to handle visibility
*/
void HandleVisibility(int visible)
{
   if (visible == GLUT_VISIBLE)
      glutIdleFunc(HandleIdle);
   else
      glutIdleFunc(NULL);
}

/*
	Create the memory for the volumetric data
	d must be either 8, 16, or 32
	Don't bother trying to recover if it fails
*/
void VolMemory(int d,int nx,int ny,int nz)
{
	int i,j;
	int failed = FALSE;

	switch (d) {
   case 8:
      if ((volume.data1 = malloc(nx*sizeof(unsigned char **))) == NULL)
			failed = TRUE;
      break;
   case 16:
      if ((volume.data2 = malloc(nx*sizeof(short **))) == NULL)
			failed = TRUE;
      break;
   case 32:
      if ((volume.data3 = malloc(nx*sizeof(int **))) == NULL)
         failed = TRUE;
      break;
   }
   if (failed) {
      fprintf(stderr,"Failed to malloc volume\n");
      CleanExit();
   }

   for (i=0;i<volume.nx;i++) {
   	switch (d) {
   	case 8:
   	   if ((volume.data1[i] = malloc(ny*sizeof(unsigned char *))) == NULL)
				failed = TRUE;
   	   break;
   	case 16:
   	   if ((volume.data2[i] = malloc(ny*sizeof(short *))) == NULL)
				failed = TRUE;
   	   break;
      case 32:
         if ((volume.data3[i] = malloc(ny*sizeof(int *))) == NULL)
            failed = TRUE;
         break;
   	}
      if (failed) {
         fprintf(stderr,"Failed to malloc plane %d\n",i);
         CleanExit();
      }
   }
   for (i=0;i<volume.nx;i++) {
      for (j=0;j<volume.ny;j++) {
      	switch (d) {
      	case 8:
      	   if ((volume.data1[i][j] = malloc(nz*sizeof(unsigned char))) == NULL)
					failed = TRUE;
      	   break;
      	case 16:
      	   if ((volume.data2[i][j] = malloc(nz*sizeof(short))) == NULL)
					failed = TRUE;
      	   break;
         case 32:
            if ((volume.data3[i][j] = malloc(nz*sizeof(int))) == NULL)
               failed = TRUE;
            break;
      	}
         if (failed) {
            fprintf(stderr,"Failed to malloc column %d\n",j);
            CleanExit();
         }
      }
   }
}

/*
	Read colour map
	Each line is expected to have 5 number, the colour index (0..65535),
	followed by the colour/alpha value.
	Gaps in the index will result in interpolation from the last value.
	Initial values are all black and perfectly transparent
	flag = 0 updates both colour and opacity, 1 update just colour, 2 update just opacity
*/
int ReadColourMap(char *fname,int flag,RGBAI *cmap)
{
   int i,j,n,ilast;
   FILE *fptr;
   const char *fn;
   RGBAI c;
   double mu;

	if (strlen(fname) < 1) {
   	fn = fl_show_file_selector("Colour and opacity map file","","","");
   	if (fn == NULL || strlen(fn) <= 0)
   	   return(FALSE);
   	if ((fptr = fopen(fn,"r")) == NULL) {
   	   fprintf(stderr,"Unable to open map file\n");
   	   return(FALSE);
   	}
	} else {
      if ((fptr = fopen(fname,"r")) == NULL) {
         fprintf(stderr,"Unable to open map file\n");
         return(FALSE);
      }
   }

	/*
		If flag is 0 then initialse to black + transparent
		Else it is assumed the colour map already exists
	*/
	if (flag == 0) {
   	for (i=0;i<MAPPINGSIZE;i++) {
      	cmap[i].r = 0;
			cmap[i].g = 0;
			cmap[i].b = 0;
			cmap[i].a = 0;
		}
	} 

   ilast = -1;
   while (fscanf(fptr,"%d %lf %lf %lf %lf",&n,&c.r,&c.g,&c.b,&c.a) == 5) {
      if (n < 0 || n >= MAPPINGSIZE) {
         fprintf(stderr,"Illegal colour index (%d), ignoring\n",n);
         continue;
      }
      if (n < ilast) {
         fprintf(stderr,"Colour index not in sequential order (%d), ignoring\n",n);
         continue;
      } else if (n == ilast) {
         fprintf(stderr,"Repeated colour index (%d), ignoring\n",n);
         continue;
      } else if (n == ilast+1) {
			if (flag == 0) {
         	cmap[n] = c;
			} else if (flag == 1) {
				cmap[n].r = c.r;
            cmap[n].g = c.g;
            cmap[n].b = c.b;
			} else if (flag == 2) {
				cmap[n].a = c.a;
			}
      } else { // Linearly interpolate
         if (ilast < 0) // Starting at other than 0
            ilast = 0;
         for (j=ilast+1;j<=n;j++) {
            mu = (j - ilast) / (double)(n - ilast);
				if (flag != 2) {
            	cmap[j].r = cmap[ilast].r + mu * (c.r - cmap[ilast].r);
            	cmap[j].g = cmap[ilast].g + mu * (c.g - cmap[ilast].g);
            	cmap[j].b = cmap[ilast].b + mu * (c.b - cmap[ilast].b);
				}
				if (flag != 1)
            	cmap[j].a = cmap[ilast].a + mu * (c.a - cmap[ilast].a);
         }
      }
      ilast = n;
   }
   fclose(fptr);
	return(TRUE);
}

/*
	Save a colour map file
	Don't write duplicates, reading will interpolate across them
*/
void SaveColourMap(void)
{
   int i,same = TRUE;
   FILE *fptr;
   const char *fn;
   char fname[128];
	RGBAI clast = {-1,-1,-1,-1};

   fn = fl_show_file_selector("Colour/opacity file","","","untitled.map");
   if (fn == NULL || strlen(fn) <= 0)
      return;
   strcpy(fname,fn);
   if ((fptr = fopen(fname,"w")) == NULL) {
      fprintf(stderr,"Failed to open map file\n");
      return;
   }

   for (i=0;i<MAPPINGSIZE;i++) {

		same = TRUE;
		if (ABS(colourmap[i].r-clast.r) > 0.001)
			same = FALSE;
      if (ABS(colourmap[i].g-clast.g) > 0.001)
         same = FALSE;
      if (ABS(colourmap[i].b-clast.b) > 0.001)
         same = FALSE;
      if (ABS(colourmap[i].a-clast.a) > 0.001)
         same = FALSE;

		if (i == 0 || i == MAPPINGSIZE-1 || !same) {
      	fprintf(fptr,"%d %g %g %g %g\n",i,
         	colourmap[i].r,colourmap[i].g,colourmap[i].b,colourmap[i].a);
			clast = colourmap[i];
		}
	}

   fclose(fptr);
}

/*
   Read a camera file
*/
int ReadCameraFile(char *fname,CAMERA *cam)
{
   FILE *fptr;
   const char *fn;

   if (strlen(fname) < 1) {
      fn = fl_show_file_selector("Camera file","","","");
      if (fn == NULL || strlen(fn) <= 0)
         return(FALSE);
      if ((fptr = fopen(fn,"r")) == NULL) {
         fprintf(stderr,"Unable to open camera file\n");
         return(FALSE);
      }
   } else {
      if ((fptr = fopen(fname,"r")) == NULL) {
         fprintf(stderr,"Unable to open camera file\n");
         return(FALSE);
      }
   }
   fscanf(fptr,"%lf %lf %lf\n",&(cam->vp.x),&(cam->vp.y),&(cam->vp.z));
   fscanf(fptr,"%lf %lf %lf\n",&(cam->vd.x),&(cam->vd.y),&(cam->vd.z));
   fscanf(fptr,"%lf %lf %lf\n",&(cam->vu.x),&(cam->vu.y),&(cam->vu.z));
   fscanf(fptr,"%lf %lf %lf\n",&(cam->pr.x),&(cam->pr.y),&(cam->pr.z));

   fscanf(fptr,"%lf\n",&(cam->focallength));
   fscanf(fptr,"%lf\n",&(cam->aperture));
   fscanf(fptr,"%lf\n",&(cam->eyesep));

   fscanf(fptr,"%d\n",&(cam->tracking));
   fscanf(fptr,"%d %d %d\n",&(cam->x0),&(cam->y0),&(cam->z0));
   fscanf(fptr,"%d %d %d\n",&(cam->xdim),&(cam->ydim),&(cam->zdim));
	MallocTextures();

   fclose(fptr);
	return(TRUE);
}

/*
	Save a camera file
*/
void SaveCameraFile(void)
{
   FILE *fptr;
   const char *fn;
   char fname[128];

   fn = fl_show_file_selector("Camera file","","","untitled.cam");
   if (fn == NULL || strlen(fn) <= 0)
      return;
   strcpy(fname,fn);
   if ((fptr = fopen(fname,"w")) == NULL) {
      fprintf(stderr,"Failed to open camera file\n");
      return;
   }
   fprintf(fptr,"%g %g %g\n",camera.vp.x,camera.vp.y,camera.vp.z);
   fprintf(fptr,"%g %g %g\n",camera.vd.x,camera.vd.y,camera.vd.z);
   fprintf(fptr,"%g %g %g\n",camera.vu.x,camera.vu.y,camera.vu.z);
   fprintf(fptr,"%g %g %g\n",camera.pr.x,camera.pr.y,camera.pr.z);
   fprintf(fptr,"%g\n",camera.focallength);
   fprintf(fptr,"%g\n",camera.aperture);
   fprintf(fptr,"%g\n",camera.eyesep);
	
	fprintf(fptr,"%d\n",camera.tracking);
	fprintf(fptr,"%d %d %d\n",camera.x0,camera.y0,camera.z0);
	fprintf(fptr,"%d %d %d\n",camera.xdim,camera.ydim,camera.zdim);
   fclose(fptr);
}

/*
   Save histogram
*/
void SaveHistogram(void)
{
	int i;
   FILE *fptr;
   const char *fn;
   char fname[128];

   fn = fl_show_file_selector("Histogram","","","untitled.hist");
   if (fn == NULL || strlen(fn) <= 0)
      return;
   strcpy(fname,fn);
   if ((fptr = fopen(fname,"w")) == NULL) {
      fprintf(stderr,"Failed to open histogram file\n");
      return;
   }
	for (i=0;i<MAPPINGSIZE;i++) {
		if (histogram[i] > 0)
			fprintf(fptr,"%d %d\n",i,histogram[i]);
	}
   fclose(fptr);
}

/*
	Calculate the integer colours from the float point ones
*/
void UpdateColourMap(void)
{
	int i;

   /* Correct bounds */
   for (i=0;i<MAPPINGSIZE;i++) {
      if (colourmap[i].r > 1) colourmap[i].r = 1;
      if (colourmap[i].r < 0) colourmap[i].r = 0;
      if (colourmap[i].g > 1) colourmap[i].g = 1;
      if (colourmap[i].g < 0) colourmap[i].g = 0;
      if (colourmap[i].b > 1) colourmap[i].b = 1;
      if (colourmap[i].b < 0) colourmap[i].b = 0;
      if (colourmap[i].a > 1) colourmap[i].a = 1;
      if (colourmap[i].a < 0) colourmap[i].a = 0;
   }

	for (i=0;i<MAPPINGSIZE;i++) {
      colourmap[i].c.r = 255 * colourmap[i].r;
      colourmap[i].c.g = 255 * colourmap[i].g;
      colourmap[i].c.b = 255 * colourmap[i].b;
      colourmap[i].c.a = 255 * colourmap[i].a;
	}
}

/*
	Calculate the lowest power of 2 that is greater than or equal to
	all the three input numbers 
*/
int CalcPower2(int i1,int i2,int i3)
{
	int two = 1;

	while (two < i1 || two < i1 || two < i3) 
		two *= 2;
	if (two > MAXTEXTURE) {
		fprintf(stderr,"CalcPower2 - this shoud not happen\n");
		two = MAXTEXTURE;
	}
	return(two);
}

/*
	Set up settings for the next frame
	Rather limited at the moment
*/
void AnimateNextFrame(int first)
{
	int i;
	static int keyframe = 1;
	static CAMERA cam1,cam2;
	static double mu1 = 0,mu2 = 0;
	static RGBAI *cmap1 = NULL,*cmap2 = NULL;
	char fname[64];

	if (first) {
		if (options.debug)
			fprintf(stderr,"Initialising animation\n");
		mu1 = 0;
		mu2 = 0;
		keyframe = 1;
		sprintf(fname,"%d.cam",keyframe);
		if (!ReadCameraFile(fname,&cam1)) {
			options.animate = FALSE;
			return;
		}
		sprintf(fname,"%d.cam",keyframe+1);
      if (!ReadCameraFile(fname,&cam2)) {
         options.animate = FALSE;
         return;
      }
		cmap1 = realloc(cmap1,MAPPINGSIZE*sizeof(RGBAI));
		sprintf(fname,"%d.map",keyframe);
		if (!ReadColourMap("1.map",0,cmap1)) {
         options.animate = FALSE;
         return;
      }
		cmap2 = realloc(cmap2,MAPPINGSIZE*sizeof(RGBAI));
		sprintf(fname,"%d.map",keyframe+1);
		if (!ReadColourMap(fname,0,cmap2)) {
         options.animate = FALSE;
         return;
      }
		return;
	}

   if (options.debug)
      fprintf(stderr,"   keyframe: %d   mu: %g\n",keyframe,mu1);

	/* Update things */
	camera.vp.x = cam1.vp.x + mu2 * (cam2.vp.x - cam1.vp.x);
   camera.vp.y = cam1.vp.y + mu2 * (cam2.vp.y - cam1.vp.y);
   camera.vp.z = cam1.vp.z + mu2 * (cam2.vp.z - cam1.vp.z);
   camera.vd.x = cam1.vd.x + mu2 * (cam2.vd.x - cam1.vd.x);
   camera.vd.y = cam1.vd.y + mu2 * (cam2.vd.y - cam1.vd.y);
   camera.vd.z = cam1.vd.z + mu2 * (cam2.vd.z - cam1.vd.z);
   camera.vu.x = cam1.vu.x + mu2 * (cam2.vu.x - cam1.vu.x);
   camera.vu.y = cam1.vu.y + mu2 * (cam2.vu.y - cam1.vu.y);
   camera.vu.z = cam1.vu.z + mu2 * (cam2.vu.z - cam1.vu.z);

	camera.x0 = cam1.x0 + mu1 * (cam2.x0 - cam1.x0);
   camera.y0 = cam1.y0 + mu1 * (cam2.y0 - cam1.y0);
   camera.z0 = cam1.z0 + mu1 * (cam2.z0 - cam1.z0);
   camera.xdim = cam1.xdim + mu1 * (cam2.xdim - cam1.xdim);
   camera.ydim = cam1.ydim + mu1 * (cam2.ydim - cam1.ydim);
   camera.zdim = cam1.zdim + mu1 * (cam2.zdim - cam1.zdim);

	for (i=0;i<MAPPINGSIZE;i++) {
		colourmap[i].r = cmap1[i].r + mu1 * (cmap2[i].r - cmap1[i].r);
      colourmap[i].g = cmap1[i].g + mu1 * (cmap2[i].g - cmap1[i].g);
      colourmap[i].b = cmap1[i].b + mu1 * (cmap2[i].b - cmap1[i].b);
      colourmap[i].a = cmap1[i].a + mu1 * (cmap2[i].a - cmap1[i].a);
	}
	UpdateColourMap();
	
	/* Value of mu for next time */
	mu1 += 1.0 / options.tweenframes;
	mu2 = 0.5*(1 - cos(mu1*PI));
	if (mu1 > 1.000001) {
      mu1 = 0;
		mu2 = 0;
      keyframe++;

      sprintf(fname,"%d.cam",keyframe);
      if (!ReadCameraFile(fname,&cam1)) {
         options.animate = FALSE;
         return;
      }
      sprintf(fname,"%d.cam",keyframe+1);
      if (!ReadCameraFile(fname,&cam2)) {
         options.animate = FALSE;
         return;
      }
      sprintf(fname,"%d.map",keyframe);
      if (!ReadColourMap(fname,0,cmap1)) {
         options.animate = FALSE;
         return;
      }
      sprintf(fname,"%d.map",keyframe+1);
      if (!ReadColourMap(fname,0,cmap2)) {
         options.animate = FALSE;
         return;
      }
	}

}







