#include "stdio.h"
#include "stdlib.h"
#include "math.h"

/*
	Simple generator of points on the Lorenz attractor
	dx/dt = a (y - x)
	dy/dt = x (b - z) - y
	dz/dt = x y - c z
	Map velocity to colour and draw as lines
*/

typedef struct {
	double r,g,b;
} COLOUR;
typedef struct {
	double x,y,z;
} XYZ;

// Generate this many points
#define N 20000

// Prototypes
COLOUR GetColour(double,double,double);
double VectorLength(XYZ,XYZ);

int main(int argc,char **argv)
{
	int i=0;
	XYZ p0 = {0.1,0,0},p1;
	COLOUR colour;
	double dp,dpmin=1e32,dpmax=-1e32;

	// Standard parameter set
	double h = 0.003;
	double a = 10.0;
	double b = 28.0;
	double c = 8.0 / 3.0;

	// Determine the range of "speeds"
	for (i=0;i<N;i++) {
      p1.x = p0.x + h * a * (p0.y - p0.x);
      p1.y = p0.y + h * (p0.x * (b - p0.z) - p0.y);
      p1.z = p0.z + h * (p0.x * p0.y - c * p0.z);
		if (i > 100) {
			dp = VectorLength(p0,p1);
			if (dp < dpmin)
				dpmin = dp;
			if (dp > dpmax)
				dpmax = dp;
		}
		p0 = p1;
	}

	// Iterate drawing an lines in GEOM f rmat
	for (i=0;i<N+100;i++) {
  		p1.x = p0.x + h * a * (p0.y - p0.x);
  		p1.y = p0.y + h * (p0.x * (b - p0.z) - p0.y);
  		p1.z = p0.z + h * (p0.x * p0.y - c * p0.z);

		// Write out the points in GEOM format
		dp = VectorLength(p0,p1);
		colour = GetColour(dp,dpmin,dpmax/2);
		printf("l %g %g %g %g %g %g %g %g %g\n",
			p0.x,p0.y,p0.z,
			p1.x,p1.y,p1.z,
			colour.r,colour.g,colour.b);

      p0 = p1;
	}

	exit(0);
}

/*
	Mapping to a blue -> red colour map in RGB sace
*/
COLOUR GetColour(double v,double vmin,double vmax)
{
	double dv;
	COLOUR c;

	// Check for weirdness
   if (vmax < vmin) {
      dv = vmin;
      vmin = vmax;
      vmax = dv;
   }
   if (vmax - vmin < 0.000001) {
      vmin -= 1;
      vmax += 1;
   }

	// Clamp values to the range
   if (v < vmin)
      v = vmin;
   if (v > vmax)
      v = vmax;
   dv = vmax - vmin;

	// Create ramps in 1/4 range segments
   if (v < (vmin + 0.25 * dv)) {
      c.r = 0;
      c.g = 4 * (v - vmin) / dv;
      c.b = 1;
   } else if (v < (vmin + 0.5 * dv)) {
      c.r = 0;
      c.g = 1;
      c.b = 1 + 4 * (vmin + 0.25 * dv - v) / dv;
   } else if (v < (vmin + 0.75 * dv)) {
      c.r = 4 * (v - vmin - 0.5 * dv) / dv;
      c.g = 1;
      c.b = 0;
   } else {
      c.r = 1;
      c.g = 1 + 4 * (vmin + 0.75 * dv - v) / dv;
      c.b = 0;
   }

	return(c);
}

double VectorLength(XYZ p1,XYZ p2)
{
   XYZ d;

   d.x = p1.x - p2.x;
   d.y = p1.y - p2.y;
   d.z = p1.z - p2.z;

   return(sqrt(d.x*d.x + d.y*d.y + d.z*d.z));
}

