#include "glgraph.h"
#include "helpstring.h"

/*
	General grid rendering
	History
	Feb 2000   Version 2, includes variable axis spacing, see readgrid() for details
              Added colour map menu
   Mar 2000   Added triangulation of random points
              Fixed sort/duplicate bug
   Apr 2000   Change to qsort() for sorting
              Improved efficiency of coincident testing
              The slowest part is now the triangulation....as it should!
   May 2000   Added GEOM file saving for stereo input, later removed
              Added point reading - without triangulation
              Rearranged the menu structure
   Jul 2000   General code cleanup
              Added general help info display
   Oct 2002   Changed to correct stereo projection
              Dropped the resolution when dragging with the mouse
              Changed handleIdle to HandleTimer as per recent conventions
              Mapped function keys to match stereo2
              Added construction lines and better lights
              Added axes grid
              Added lots of contouring options
   Nov 2006   General cleanup and rewrite, created OPTIONS structure
              Made makefile and simplified Mac/Linux builds
              Added SPIKE and BOX data types, rewrote MakeGeometry()
	To do:
   Add axes labels
*/

// Form library 
#include "stereoform.c"
FD_stereoform *stereoform;
int stereoformvisible = FALSE;
#include "scaleform.c"
FD_scaleform *scaleform;
int scaleformvisible = FALSE;
#include "contourform.c"
FD_contourform *contourform;
int contourformvisible = TRUE;

// Geometry Structure 
int nx = 0;
int ny = 0; 
XYZ *grid = NULL; // Acts as a 2D grid or list of vertices for points and triangulation
int ntri = 0;
ITRIANGLE *triangles = NULL;
XYZ pmin={-1.0,-1.0,-1.0},pmax={1.0,1.0,1.0},pmid;

// Options
OPTIONS options;
double framerate = 0;

// Camera 
CAMERA camera;
XYZ thelights[2];

int main(int argc,char **argv)
{
   int i;
	char fname[128],viewname[128];
	char s[64];

	options.stereo           = NOSTEREO;
	options.combineeyes      = FALSE;
	options.fullscreen       = FALSE;
	options.geommode         = SURFACE;
	options.rendermode       = WIREFRAME;
	options.showhelpinfo     = FALSE;
	options.colourmap        = 1;
	options.constcolour.r    = 0.5; options.constcolour.g = 0.5; options.constcolour.b = 0.5;
	options.showaxes         = TRUE;
	options.showaxesgrid     = TRUE;
	options.showboundingbox  = FALSE;
	options.showconstruction = FALSE;
	options.resolution       = 1;
	options.debug            = FALSE;
	options.currentbutton    = -1;
	options.scalefactors.x   = 1; options.scalefactors.y = 1; options.scalefactors.z = 1;
	options.dobigdump        = FALSE;
	options.windowdump       = FALSE;
	options.recordimages     = FALSE;
	options.ncontour         = 40;
	options.targetfps        = 30;
	options.background.r     = 0; options.background.g = 0; options.background.b = 0;
	options.datatype         = NODATA;  // Either: NODATA, GRIDDED, TRIANGULAR, POINTED, GRIDDED2 

	// Animation settings 
	options.autopilot = FALSE;
	options.autospin = FALSE;
	options.dtheta = 1;

   for (i=1;i<argc;i++) {
      if (strcmp(argv[i],"-h") == 0) 
         GiveUsage(argv[0]);
      if (strcmp(argv[i],"-d") == 0)
         options.debug = TRUE;

		if (strcmp(argv[i],"-s") == 0)
			options.stereo = ACTIVESTEREO;
		if (strcmp(argv[i],"-ss") == 0)
         options.stereo = DUALSTEREO;
      if (strcmp(argv[i],"-f") == 0)
         options.fullscreen = TRUE;

		if (strcmp(argv[i],"-rp") == 0)
         options.geommode = POINTS;
      if (strcmp(argv[i],"-rs") == 0) 
         options.geommode = SURFACE;
      if (strcmp(argv[i],"-rk") == 0)
         options.geommode = SPIKES;
      if (strcmp(argv[i],"-rb") == 0)
			options.geommode = BOXES;
      if (strcmp(argv[i],"-rx") == 0)
         options.geommode = XCONTOUR;
      if (strcmp(argv[i],"-ry") == 0)
         options.geommode = YCONTOUR;
      if (strcmp(argv[i],"-rz") == 0)
         options.geommode = ZCONTOUR;
      if (strcmp(argv[i],"-rxy") == 0)
         options.geommode = XYCONTOUR;

      if (strcmp(argv[i],"-sw") == 0)
         options.rendermode = WIREFRAME;
      if (strcmp(argv[i],"-sf") == 0)
         options.rendermode = FLAT;
      if (strcmp(argv[i],"-ss") == 0)
         options.rendermode = SMOOTH;
		if (strcmp(argv[i],"-sl") == 0L)
         options.rendermode = LIGHT;

      if (strcmp(argv[i],"-z") == 0) {
         i++;
         if (i >= argc) {
            GiveUsage(argv[0]);
            continue;
         }
         options.scalefactors.y = atof(argv[i]);
      }

		if (strcmp(argv[i],"-og") == 0) {
			i++;
			if (i >= argc) {
				GiveUsage(argv[0]);
				continue;
			}
			strcpy(fname,argv[i]);
			options.datatype = GRIDDED;
		}
      if (strcmp(argv[i],"-og2") == 0) {
         i++;
         if (i >= argc) {
            GiveUsage(argv[0]);
            continue;
         }
         strcpy(fname,argv[i]);
         options.datatype = GRIDDED2;
      }
      if (strcmp(argv[i],"-or") == 0) {
         i++;
         if (i >= argc) {
            GiveUsage(argv[0]);
            continue;
         }
         strcpy(fname,argv[i]);
			options.datatype = TRIANGULAR;
      }
      if (strcmp(argv[i],"-op") == 0) {
         i++;
         if (i >= argc) {
            GiveUsage(argv[0]);
            continue;
         }
         strcpy(fname,argv[i]);
			options.datatype = POINTED;
      }

      if (strcmp(argv[i],"-vf") == 0) {
         i++;
         if (i >= argc) {
            GiveUsage(argv[0]);
            continue;
         }
         strcpy(viewname,argv[i]);
      }
      if (strcmp(argv[i],"-res") == 0) {
         i++;
         if (i >= argc) {
            GiveUsage(argv[0]);
            continue;
         }
         options.resolution = atoi(argv[i]);
      }
   }

   // Set things up and go 
   camera.screenwidth  = 800;
   camera.screenheight = 600;
	if (options.stereo == DUALSTEREO)
		camera.screenwidth *= 2;
   glutInit(&argc,argv);
	if (options.stereo == ACTIVESTEREO)
		glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STEREO);
	else
   	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutCreateWindow("3D stereo graphs");
	glutReshapeWindow(camera.screenwidth,camera.screenheight);
	if (options.fullscreen)
		glutFullScreen();
   glutDisplayFunc(HandleDisplay);
	glutReshapeFunc(HandleReshape);
   glutVisibilityFunc(HandleVisibility);
   glutKeyboardFunc(HandleKeyboard);
	glutIdleFunc(HandleIdle);
   glutSpecialFunc(HandleSpecialKeyboard);
   glutMouseFunc(HandleMouse);
	glutMotionFunc(HandleMouseMotion);
	glutPassiveMotionFunc(HandlePassiveMotion);
	/* glutSetCursor(GLUT_CURSOR_NONE); */
   CreateEnvironment();
	CreateMenus();

   /* Create the stereo forms */
	i = 1;
   fl_initialize(&i,argv,"Stereographic grapher",0,0);
   stereoform = create_form_stereoform();
   fl_set_slider_bounds(stereoform->Eyesep,0.0,camera.focallength/2);
   fl_set_slider_bounds(stereoform->Aperture,0.0,90.0);
   fl_set_slider_bounds(stereoform->Focallength,camera.focallength/2,4*camera.focallength);

	/* Create the scale form */
	scaleform = create_form_scaleform();
	sprintf(s,"%g",options.scalefactors.y);
   fl_set_input(scaleform->ZScale,s);

   /* Create the contour form */
   contourform = create_form_contourform();
   sprintf(s,"%d",options.ncontour);
   fl_set_input(contourform->Ncontour,s);

	/* Initial model if there is one */
	if (options.datatype == GRIDDED) {
      ReadGrid(fname);
	} else if (options.datatype == GRIDDED2) {
      ReadGrid(fname);
	} else if (options.datatype == TRIANGULAR) {
		ReadRandom(fname,TRUE);
	} else if (options.datatype == POINTED) {
		ReadRandom(fname,FALSE);
	}
	CameraHome(HOME);
   MakeGeometry();

	/* Read the view file if there is one */
	if (strlen(viewname) > 0)
  		ReadViewFile(viewname,&camera);

   UpdateForm();
   glutMainLoop();
   return(0);
}

/*
   This is where global settings are made, 
   things that will not change per frame
*/
void CreateEnvironment(void)
{
   GLint num[10];
   GLfloat fsizes[2],fstep;

   glEnable(GL_DEPTH_TEST);
   glEnable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_DITHER);
   glDisable(GL_LINE_SMOOTH);
   glDisable(GL_POINT_SMOOTH);
   glDisable(GL_POLYGON_SMOOTH); 
   glLineWidth(1.0);
   glPointSize(1.0);
   glFrontFace(GL_CCW);
   glDisable(GL_CULL_FACE);
   glClearColor(options.background.r,options.background.g,options.background.b,0.0); 
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);

	glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
   glEnable(GL_COLOR_MATERIAL);
   glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_FALSE);
   glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);

   /* Test miscellaneous aspects of the OpenGL state */
   if (options.debug) {
      glGetFloatv(GL_POINT_SIZE_RANGE,fsizes);
      glGetFloatv(GL_POINT_SIZE_GRANULARITY,&fstep);
      fprintf(stderr,"Point sizes supported: %g to %g in steps of %g\n",
         fsizes[0],fsizes[1],fstep);
      glGetFloatv(GL_LINE_WIDTH_RANGE,fsizes);
      glGetFloatv(GL_LINE_WIDTH_GRANULARITY,&fstep);
      fprintf(stderr,"Line widths supported: %g to %g in steps of %g\n",
         fsizes[0],fsizes[1],fstep);
      glGetIntegerv(GL_MAX_PROJECTION_STACK_DEPTH,num);
      fprintf(stderr,"Projection stack depth: %d\n",num[0]);
      glGetIntegerv(GL_MAX_MODELVIEW_STACK_DEPTH,num);
      fprintf(stderr,"ModelView stack depth: %d\n",num[0]);
      glGetIntegerv(GL_MAX_CLIP_PLANES,num);
      fprintf(stderr,"Number of clipping planes: %d\n",num[0]);
      glGetIntegerv(GL_DEPTH_BITS,num);
      fprintf(stderr,"Depth buffer \"depth\": %d\n",num[0]);
      glGetIntegerv(GL_MAX_VIEWPORT_DIMS,num);
      fprintf(stderr,"Maximum viewport size: %d x %d\n",num[0],num[1]);
   }
}

/*
   Handle a window reshape/resize
*/
void HandleReshape(int w,int h)
{
   glViewport(0,0,(GLsizei)w,(GLsizei)h);
	camera.screenwidth = w;
	camera.screenheight = h;
}

/*
   This is the basic display callback routine 
	called to redraw the buffers optionally in stereo
*/
void HandleDisplay(void)
{
	int i,j;
	XYZ r;
   double ratio,radians,wd2,ndfl;
   double left,right,top,bottom,near=0.1,far=10000;
   double deltaeye;
	XYZ p1,p2;
   static int framecount = -1;
   static double tstart,tstop;         /* Frame rate calculations */

   /* Set the time the first time around */
   if (framecount < 0) {
      framecount = 0;
      tstart = GetRunTime();
   }
   framecount++;

   /* Sort out the eye separation */
   if (options.combineeyes)
      deltaeye = 0;
   else
      deltaeye = camera.eyesep / 2;

   /* Clip to avoid extreme stereo */
	near = camera.focallength / 10;
	p1 = SetPoint(pmin);
	p2 = SetPoint(pmax);
	far = 100 * VectorLength(p1,p2);
   if (options.stereo == ACTIVESTEREO || options.stereo == DUALSTEREO) 
      near = camera.focallength / 5;

   /* Misc stuff */
   ratio   = camera.screenwidth / (double)camera.screenheight;
	if (options.stereo == DUALSTEREO)
		ratio /= 2;
   radians = DTOR * camera.aperture / 2;
   wd2     = near * tan(radians);
   ndfl    = near / camera.focallength;

   /* Are we in autopilot form */
   if (options.autopilot)
      options.autopilot = AutoPilot(1);

   /* Are we in autospin mode */
   if (options.autospin) 
      RotateCamera(0.1,0.0,0.0,1.0);

	if (options.stereo == ACTIVESTEREO || options.stereo == DUALSTEREO) {

      if (options.stereo == DUALSTEREO) {
         glDrawBuffer(GL_BACK);
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      }

      /* Derive the two eye positions */
      CROSSPROD(camera.vd,camera.vu,r);
		Normalise(&r);
		r.x *= deltaeye;
      r.y *= deltaeye;
      r.z *= deltaeye;

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      left  = - ratio * wd2 - deltaeye * ndfl;
      right =   ratio * wd2 - deltaeye * ndfl;
      top    =   wd2;
      bottom = - wd2;
      glFrustum(left,right,bottom,top,near,far);
		if (options.stereo == DUALSTEREO)
			glViewport(camera.screenwidth/2,0,camera.screenwidth/2,camera.screenheight);
		else
			glViewport(0,0,camera.screenwidth,camera.screenheight);

   	glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK_RIGHT);
      if (options.stereo == ACTIVESTEREO)
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glLoadIdentity();
      gluLookAt(camera.vp.x + r.x,camera.vp.y + r.y,camera.vp.z + r.z,
                camera.vp.x + r.x + camera.vd.x,
                camera.vp.y + r.y + camera.vd.y,
                camera.vp.z + r.z + camera.vd.z,
					 camera.vu.x,camera.vu.y,camera.vu.z);
   	MakeLighting();
   	MakeMaterial();
		glCallList(1000);
		DrawExtras();
		
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      left  = - ratio * wd2 + deltaeye * ndfl;
      right =   ratio * wd2 + deltaeye * ndfl;
      top    =   wd2;
      bottom = - wd2;
      glFrustum(left,right,bottom,top,near,far);
      if (options.stereo == DUALSTEREO)
         glViewport(0,0,camera.screenwidth/2,camera.screenheight);
      else
         glViewport(0,0,camera.screenwidth,camera.screenheight);

      glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK_LEFT);
      if (options.stereo == ACTIVESTEREO)
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();
      gluLookAt(camera.vp.x - r.x,camera.vp.y - r.y,camera.vp.z - r.z,
                camera.vp.x - r.x + camera.vd.x,
                camera.vp.y - r.y + camera.vd.y,
                camera.vp.z - r.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
      MakeLighting();
      MakeMaterial();
		glCallList(1000);
		DrawExtras();

	} else {

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      left  = - ratio * wd2;
      right =   ratio * wd2;
      top    =   wd2;
      bottom = - wd2;
      glFrustum(left,right,bottom,top,near,far);
		glViewport(0,0,camera.screenwidth,camera.screenheight);

      glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK_LEFT);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();
      gluLookAt(camera.vp.x,camera.vp.y,camera.vp.z,
                camera.vp.x + camera.vd.x,
                camera.vp.y + camera.vd.y,
                camera.vp.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
      MakeLighting();
      MakeMaterial();
		glCallList(1000);
		DrawExtras();
	}

   /* 
		High res dump hack, make a 4x4 grid of images 
		Only do mono, not stereo
	*/
   if (options.dobigdump) {

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      left  = - ratio * wd2;
      right =   ratio * wd2;
      top    =   wd2;
      bottom = - wd2;
      glFrustum(left,right,bottom,top,near,far);

      for (i=0;i<4;i++) {
         for (j=0;j<4;j++) {
            fprintf(stderr,"Frame (%d,%d)\n",i,j);
            glViewport(
               -i*camera.screenwidth,-j*camera.screenheight,
               4*camera.screenwidth,4*camera.screenheight);
            glMatrixMode(GL_MODELVIEW);
            glDrawBuffer(GL_BACK);
				glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glLoadIdentity();
            gluLookAt(camera.vp.x,camera.vp.y,camera.vp.z,
                		 camera.vp.x + camera.vd.x,
                		 camera.vp.y + camera.vd.y,
                		 camera.vp.z + camera.vd.z,
                      camera.vu.x,camera.vu.y,camera.vu.z);
            MakeLighting();
            MakeMaterial();
           	glCallList(1000);
				if (options.stereo == ACTIVESTEREO)
            	WindowDump("",camera.screenwidth,camera.screenheight,TRUE,12);
				else
					WindowDump("",camera.screenwidth,camera.screenheight,FALSE,12);
				glutSwapBuffers();
         }
      }
      options.dobigdump = FALSE;
   }

	if (options.recordimages || options.windowdump) {
      if (options.stereo == ACTIVESTEREO)
         WindowDump("",camera.screenwidth,camera.screenheight,TRUE,12);
      else
         WindowDump("",camera.screenwidth,camera.screenheight,FALSE,12);
		options.windowdump = FALSE;
	}

   // glFlush();  This isn't necessary for double buffers 
   glutSwapBuffers();

   // Calculate an optionally show the frame rate 
   tstop = GetRunTime();
   if (tstop - tstart > 5) {
      framerate = framecount / (tstop - tstart);
      if (options.debug)
         fprintf(stderr,"Frame rate = %.1f frames/second\n",framerate);
      framecount = 0;
      tstart = tstop;
	}
}

/*
	Draw any extra stuff
*/
void DrawExtras(void)
{
	int y = 10;
	COLOUR colour = {0.0,0.0,0.0};
	char s[32];

   // Inverse colour 
   colour.r = 1 - options.background.r;
   colour.g = 1 - options.background.g;
   colour.b = 1 - options.background.b;
	glColor3f(colour.r,colour.g,colour.b);

	glDisable(GL_LIGHTING);

   // Are we showing text help information? 
   if (options.showhelpinfo) {
      DrawGLText(10,camera.screenheight - 20,helpstring);
      sprintf(s,"Frame rate: %.1f frames/second",framerate);
      DrawGLText(10,y+=13,s);
   }

	// Axes 
	if (options.showaxes || options.showaxesgrid)
   	DrawTheAxes();

	// Bounding box 
	if (options.showboundingbox)
   	DrawBoundingBox();

	// Lights
	if (options.showconstruction)
   	DrawLights();
}

/*
	Create graphical representations for each datatype and for each drawing mode.
	Note that some drawing mode are only relevant to certain datatypes.
	If mode is "GEOMFILE" then write geometry to "fptr"
	This is optimised for the different datatypes in an attempt to
	handle very large datasets.
*/
void MakeGeometry(void)
{
   int i,j,k,index;
	XYZ p[4],normal[4],plast = {0,0,0};
	XYZ p0,p1,p2,p3[3],n;
	COLOUR colour[4];

	// Is there anything to draw 
	//if (options.datatype == NODATA || (grid == NULL && triangles == NULL)) 
	//	return;
	if (options.debug)
		fprintf(stderr,"Updating display lists\n");

	ComputeBounds();
	glNewList(1000,GL_COMPILE);

	// =====================================================================
	// Point data
	// Spikes or point rendering allowed
   if (options.datatype == POINTED) {

		// Draw as spikes
		if (options.geommode == SPIKES) {
         glBegin(GL_LINES);
         for (i=0;i<nx*ny;i+=options.resolution) {
            colour[0] = SetColour(grid[i]);
            p[0] = SetPoint(grid[i]);
				p1 = grid[i];
				p1.y = pmin.y;
				p[1] = SetPoint(p1);
            glVertex3f(p[0].x,p[0].y,p[0].z);
            glVertex3f(p[1].x,p[1].y,p[1].z);
         }
         glEnd();
		} 

		// Points
		else if (options.geommode == POINTS) {
      	glBegin(GL_POINTS);
      	for (i=0;i<nx*ny;i+=options.resolution) {
				colour[0] = SetColour(grid[i]);
				p[0] = SetPoint(grid[i]);
      	   glVertex3f(p[0].x,p[0].y,p[0].z);
      	}
      	glEnd();
		}
   }

   // =====================================================================
	// Handle the two sorts of gridded data
	// Rendering modes, points, spikes, surfaces, contours, boxes
	if (options.datatype == GRIDDED || options.datatype == GRIDDED2) {

		// Draw as points
		if (options.geommode == POINTS) {
      	glBegin(GL_POINTS);
			for (i=0;i<nx;i+=options.resolution) {
				for (j=0;j<ny;j+=options.resolution) {
					index = j * nx + i;
  	    	   	colour[0] = SetColour(grid[index]);
      	   	p[0] = SetPoint(grid[index]);
      	   	glVertex3f(p[0].x,p[0].y,p[0].z);
				}
      	}
      	glEnd();
		}

		// Draw as spikes
      else if (options.geommode == SPIKES) {
         glBegin(GL_LINES);
         for (i=0;i<nx;i+=options.resolution) {
            for (j=0;j<ny;j+=options.resolution) {
					index = j * nx + i;
	            colour[0] = SetColour(grid[index]);
   	         p[0] = SetPoint(grid[index]);
      	      p1 = grid[index];
         	   p1.y = pmin.y;
         	   p[1] = SetPoint(p1);
         	   glVertex3f(p[0].x,p[0].y,p[0].z);
         	   glVertex3f(p[1].x,p[1].y,p[1].z);
         	}
			}
         glEnd();
		}

		// Contours
		else if (options.geommode == XCONTOUR || 
					options.geommode == YCONTOUR || 
					options.geommode == XYCONTOUR) {

			if (options.geommode == XCONTOUR || 
				 options.geommode == XYCONTOUR) {
            n.x = 1;
            n.y = 0;
            n.z = 0;
            glBegin(GL_LINES);
				for (k=0;k<=options.ncontour;k++) {
               p0.x = pmin.x + k * (pmax.x - pmin.x) / options.ncontour;
               p0.y = 0;
               p0.z = 0;
      			for (i=0;i<nx-options.resolution;i+=options.resolution) {
         			for (j=0;j<ny-options.resolution;j+=options.resolution) {
							index = j * nx + i;
                  	p3[0] = grid[index];
                  	p3[1] = grid[index + options.resolution];
                  	p3[2] = grid[index + options.resolution * nx + options.resolution];
                  	if (ContourFacet(p3,n,p0,&p1,&p2) == 2) {
                  	   colour[0] = SetColour(p1);
                  	   p1 = SetPoint(p1);
                  	   glVertex3f(p1.x,p1.y,p1.z);
                  	   colour[0] = SetColour(p2);
                  	   p2 = SetPoint(p2);
                  	   glVertex3f(p2.x,p2.y,p2.z);
                  	}
                     p3[0] = grid[index];
                     p3[1] = grid[index + options.resolution * nx + options.resolution];
							p3[2] = grid[index + options.resolution * nx];
                     if (ContourFacet(p3,n,p0,&p1,&p2) == 2) {
                        colour[0] = SetColour(p1);
                        p1 = SetPoint(p1);
                        glVertex3f(p1.x,p1.y,p1.z);
                        colour[0] = SetColour(p2);
                        p2 = SetPoint(p2);
                        glVertex3f(p2.x,p2.y,p2.z);
                     }
						}
               }
            }
            glEnd();
			}

         if (options.geommode == YCONTOUR || 
				 options.geommode == XYCONTOUR) {
            n.x = 0;
            n.y = 0;
            n.z = 1;
            glBegin(GL_LINES);
            for (k=0;k<=options.ncontour;k++) {
               p0.x = 0;
               p0.y = 0;
               p0.z = pmin.z + k * (pmax.z - pmin.z) / options.ncontour;
               for (i=0;i<nx-options.resolution;i+=options.resolution) {
                  for (j=0;j<ny-options.resolution;j+=options.resolution) {
							index = j * nx + i;
                     p3[0] = grid[index];
                     p3[1] = grid[index + options.resolution];
                     p3[2] = grid[index + options.resolution * nx + options.resolution];
                     if (ContourFacet(p3,n,p0,&p1,&p2) == 2) {
                        colour[0] = SetColour(p1);
                        p1 = SetPoint(p1);
                        glVertex3f(p1.x,p1.y,p1.z);
                        colour[0] = SetColour(p2);
                        p2 = SetPoint(p2);
                        glVertex3f(p2.x,p2.y,p2.z);
                     }
                     p3[0] = grid[index];
                     p3[1] = grid[index + options.resolution * nx + options.resolution];
                     p3[2] = grid[index + options.resolution * nx];
                     if (ContourFacet(p3,n,p0,&p1,&p2) == 2) {
                        colour[0] = SetColour(p1);
                        p1 = SetPoint(p1);
                        glVertex3f(p1.x,p1.y,p1.z);
                        colour[0] = SetColour(p2);
                        p2 = SetPoint(p2);
                        glVertex3f(p2.x,p2.y,p2.z);
                     }
                  }
               }
            }
            glEnd();
         }
		}

      else if (options.geommode == ZCONTOUR) {
         n.x = 0;
         n.y = 1;
         n.z = 0;
         glBegin(GL_LINES);
         for (k=0;k<=options.ncontour;k++) {
            p0.x = 0;
            p0.y = pmin.y + k * (pmax.y - pmin.y) / options.ncontour;
            p0.z = 0;
            for (i=0;i<nx-options.resolution;i+=options.resolution) {
               for (j=0;j<ny-options.resolution;j+=options.resolution) {
						index = j * nx + i;
                  p3[0] = grid[index];
                  p3[1] = grid[index + options.resolution];
                  p3[2] = grid[index + options.resolution * nx + options.resolution];
                  if (ContourFacet(p3,n,p0,&p1,&p2) == 2) {
                     colour[0] = SetColour(p1);
                     p1 = SetPoint(p1);
                     glVertex3f(p1.x,p1.y,p1.z);
                     colour[0] = SetColour(p2);
                     p2 = SetPoint(p2);
                     glVertex3f(p2.x,p2.y,p2.z);
                  }
                  p3[0] = grid[index];
                  p3[1] = grid[index + options.resolution * nx + options.resolution];
                  p3[2] = grid[index + options.resolution * nx];
                  if (ContourFacet(p3,n,p0,&p1,&p2) == 2) {
                     colour[0] = SetColour(p1);
                     p1 = SetPoint(p1);
                     glVertex3f(p1.x,p1.y,p1.z);
                     colour[0] = SetColour(p2);
                     p2 = SetPoint(p2);
                     glVertex3f(p2.x,p2.y,p2.z);
                  }
               }
            }
         }
         glEnd();
      }

		// Draw surface
		else if (options.geommode == SURFACE) {
      	glBegin(GL_QUADS);
      	for (i=0;i<nx-options.resolution;i+=options.resolution) {
      	   for (j=0;j<ny-options.resolution;j+=options.resolution) {
	
	            index = j * nx + i;
	            normal[0] = GridNormal(i,j);
	            glNormal3f(normal[0].x,normal[0].y,normal[0].z);
					colour[0] = SetColour(grid[index]);
					p[0] = SetPoint(grid[index]);
	            glVertex3f(p[0].x,p[0].y,p[0].z);
	
	            index += options.resolution;
	            normal[1] = GridNormal(i+options.resolution,j);
	            glNormal3f(normal[1].x,normal[1].y,normal[1].z);
					colour[1] = SetColour(grid[index]);
					p[1] = SetPoint(grid[index]);
	            glVertex3f(p[1].x,p[1].y,p[1].z);
	
	            index = (j+options.resolution) * nx + i + options.resolution;
	            normal[2] = GridNormal(i+options.resolution,j+options.resolution);
	            glNormal3f(normal[2].x,normal[2].y,normal[2].z);
					colour[2] = SetColour(grid[index]);
					p[2] = SetPoint(grid[index]);
	            glVertex3f(p[2].x,p[2].y,p[2].z);
	
	            index -= options.resolution;
	            normal[3] = GridNormal(i,j+options.resolution);
	            glNormal3f(normal[3].x,normal[3].y,normal[3].z);
					colour[3] = SetColour(grid[index]);
					p[3] = SetPoint(grid[index]);
	            glVertex3f(p[3].x,p[3].y,p[3].z);
				}
			}
			glEnd();
		}

		else if (options.geommode == BOXES) {
         for (i=0;i<=nx-options.resolution;i+=options.resolution) {
            for (j=0;j<=ny-options.resolution;j+=options.resolution) {
					index = j * nx + i;
					if (i == 0)
						p0.x = grid[j*nx+i].x - (grid[j*nx+i+1].x-grid[j*nx+i].x)/2;
					else
						p0.x = grid[j*nx+i].x - (grid[j*nx+i].x-grid[j*nx+i-1].x)/2;
					p0.y = pmin.y;
					if (j == 0)
                  p0.z = grid[j*nx+i].z - (grid[(j+1)*nx+i].z-grid[j*nx+i].z)/2;
					else
						p0.z = grid[j*nx+i].z - (grid[j*nx+i].z-grid[(j-1)*nx+i].z)/2;
					if (i == nx-1)
                  p1.x = grid[j*nx+i].x + (grid[j*nx+i].x-grid[j*nx+i-1].x)/2;
					else
                  p1.x = grid[j*nx+i].x + (grid[j*nx+i+1].x-grid[j*nx+i].x)/2;
               p1.y = grid[j*nx+i].y;
					if (j == ny-1)
                  p1.z = grid[j*nx+i].z + (grid[j*nx+i].z-grid[(j-1)*nx+i].z)/2;
					else
                  p1.z = grid[j*nx+i].z + (grid[(j+1)*nx+i].z-grid[j*nx+i].z)/2;
					p[0] = SetPoint(p0);
					p[1] = SetPoint(p1);
					colour[0] = SetColour(grid[index]);
					DrawABox(p[0],p[1]);
				}
			}
		}

      // Wireframe, default
      else {

         for (i=0;i<nx;i+=options.resolution) {
            glBegin(GL_LINE_STRIP);
            for (j=0;j<ny;j+=options.resolution) {
               index = j * nx + i;
               colour[0] = SetColour(grid[index]);
               p[0] = SetPoint(grid[index]);
               glVertex3f(p[0].x,p[0].y,p[0].z);
               plast = p[0];
            }
            glEnd();
         }
         for (j=0;j<ny;j+=options.resolution) {
            glBegin(GL_LINE_STRIP);
            for (i=0;i<nx;i+=options.resolution) {
               index = j * nx + i;
               colour[0] = SetColour(grid[index]);
               p[0] = SetPoint(grid[index]);
               glVertex3f(p[0].x,p[0].y,p[0].z);
               plast = p[0];
            }
            glEnd();
         }
      }
	}

	// =====================================================================
	// Handle the triangulated data 
	// Rendering modes: points, spikes, contours, surfaces, wireframe
	if (options.datatype == TRIANGULAR) {

      // Draw as points
      if (options.geommode == POINTS) {
         glBegin(GL_POINTS); 
         for (i=0;i<nx;i+=options.resolution) {
            colour[0] = SetColour(grid[i]);
            p[0] = SetPoint(grid[i]);
            glVertex3f(p[0].x,p[0].y,p[0].z);
         }     
         glEnd(); 
      }              

		// Draw as spikes
      else if (options.geommode == SPIKES) {
         glBegin(GL_LINES);
         for (i=0;i<nx;i+=options.resolution) {
            colour[0] = SetColour(grid[i]);
            p[0] = SetPoint(grid[i]);
            p1 = grid[i];
            p1.y = pmin.y;
            p[1] = SetPoint(p1);
            glVertex3f(p[0].x,p[0].y,p[0].z);
            glVertex3f(p[1].x,p[1].y,p[1].z);
         }
         glEnd();
      }     

		// Contours
		else if (options.geommode == XCONTOUR || 
					options.geommode == YCONTOUR || 
					options.geommode == XYCONTOUR) {

			if (options.geommode == XCONTOUR || 
				 options.geommode == XYCONTOUR) {
				n.x = 1;
				n.y = 0;
				n.z = 0;
				glBegin(GL_LINES);
				for (j=0;j<=options.ncontour;j++) {
               p0.x = pmin.x + j * (pmax.x - pmin.x) / options.ncontour;
               p0.y = 0;
               p0.z = 0;
					for (i=0;i<ntri;i++) {
						p3[0] = grid[triangles[i].p1];
						p3[1] = grid[triangles[i].p2];
						p3[2] = grid[triangles[i].p3];
						if (ContourFacet(p3,n,p0,&p1,&p2) == 2) {
							colour[0] = SetColour(p1);
							p1 = SetPoint(p1);
							glVertex3f(p1.x,p1.y,p1.z);	
							colour[0] = SetColour(p2);
							p2 = SetPoint(p2);
							glVertex3f(p2.x,p2.y,p2.z);
						}
					}
				}
				glEnd();
			}

      	if (options.geommode == YCONTOUR || 
				 options.geommode == XYCONTOUR) {
         	n.x = 0;
         	n.y = 0;
         	n.z = 1;
         	glBegin(GL_LINES);
            for (j=0;j<=options.ncontour;j++) {
               p0.x = 0;
               p0.y = 0;
               p0.z = pmin.z + j * (pmax.z - pmin.z) / options.ncontour;
         		for (i=0;i<ntri;i++) {
         	      p3[0] = grid[triangles[i].p1];
         	      p3[1] = grid[triangles[i].p2];
         	      p3[2] = grid[triangles[i].p3];
         	      if (ContourFacet(p3,n,p0,&p1,&p2) == 2) {
         	         colour[0] = SetColour(p1);
         	         p1 = SetPoint(p1);
         	         glVertex3f(p1.x,p1.y,p1.z);
         	         colour[0] = SetColour(p2);
         	         p2 = SetPoint(p2);
         	         glVertex3f(p2.x,p2.y,p2.z);
         	      }
         	   }
         	}
         	glEnd();
			}
      }

      else if (options.geommode == ZCONTOUR) {
         n.x = 0;
         n.y = 1;
         n.z = 0;
         glBegin(GL_LINES);
         for (j=0;j<=options.ncontour;j++) {
            p0.x = 0;
            p0.y = pmin.y + j * (pmax.y - pmin.y) / options.ncontour;
            p0.z = 0;
         	for (i=0;i<ntri;i++) {
               p3[0] = grid[triangles[i].p1];
               p3[1] = grid[triangles[i].p2];
               p3[2] = grid[triangles[i].p3];
               if (ContourFacet(p3,n,p0,&p1,&p2) == 2) {
                  colour[0] = SetColour(p1);
                  p1 = SetPoint(p1);
                  glVertex3f(p1.x,p1.y,p1.z);
                  colour[0] = SetColour(p2);
                  p2 = SetPoint(p2);
                  glVertex3f(p2.x,p2.y,p2.z);
               }
            }
         }
         glEnd();
      }

		// Draw surface
		else if (options.geommode == SURFACE) {

			glBegin(GL_TRIANGLES);
   		for (i=0;i<ntri;i++) {
      	   normal[0] = CalcNormal(grid[triangles[i].p1],grid[triangles[i].p3],grid[triangles[i].p2]);
				normal[1] = normal[0];
				normal[2] = normal[0];
	
				index = triangles[i].p1;
				glNormal3f(normal[0].x,normal[0].y,normal[0].z);
				colour[0] = SetColour(grid[index]);
				p[0] = SetPoint(grid[index]);
	      	glVertex3f(p[0].x,p[0].y,p[0].z);
	
	      	index = triangles[i].p2;
	         glNormal3f(normal[1].x,normal[1].y,normal[1].z);
				colour[1] = SetColour(grid[index]);
				p[1] = SetPoint(grid[index]);
	      	glVertex3f(p[1].x,p[1].y,p[1].z);
	
	      	index = triangles[i].p3;
	         glNormal3f(normal[2].x,normal[2].y,normal[2].z);
				colour[2] = SetColour(grid[index]);
				p[2] = SetPoint(grid[index]);
	      	glVertex3f(p[2].x,p[2].y,p[2].z);
			}
			glEnd();
		}

	}

	glEndList();
	glEnd();
}

/*
	Draw the axes
	Possibly with axes plane grids
*/
void DrawTheAxes(void)
{
	int i,n=10;
	double x,y,z,dx,dy,dz,dd;
	COLOUR border = {0.5,0.5,0.5};

	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);
	
	// The length of the tick marks 
	dx = (pmax.x - pmin.x) / 40;
	dy = options.scalefactors.y / 50;
	dz = (pmax.z - pmin.z) / 40;
	dd = MAX(dx,dz);

	if (options.showaxes) {

   	/* X axis - red */
   	glColor3f(1.0,0.0,0.0);
   	glBegin(GL_LINES);
   	glVertex3f(pmin.x,0.0,pmin.z);
   	glVertex3f(pmax.x+dx,0.0,pmin.z);
		for (i=0;i<=n;i++) {
			x = pmin.x + i * (pmax.x -pmin.x) / n;
			glVertex3f(x,-dd,pmin.z);
   		glVertex3f(x,0.0,pmin.z);	
		}
		glEnd();

   	/* Z axis - blue */
   	glColor3f(0.0,0.0,1.0);
   	glBegin(GL_LINES);
   	glVertex3f(pmin.x,0.0,pmin.z);
   	glVertex3f(pmin.x,0.0,pmax.z+dz);
   	for (i=0;i<=n;i++) {
   	   z = pmin.z + i * (pmax.z - pmin.z) / n;
   	   glVertex3f(pmin.x,-dd,z);
   	   glVertex3f(pmin.x,0.0,z);   
   	}
   	glEnd();

   	/* Y axis  - green */
   	glColor3f(0.0,1.0,0.0);
   	glBegin(GL_LINES);
   	glVertex3f(pmin.x,0.0,pmin.z);
   	glVertex3f(pmin.x,options.scalefactors.y+dy,pmin.z);
      for (i=0;i<=n;i++) {
         y = i * options.scalefactors.y / n;
         glVertex3f(pmin.x,y,pmin.z);
         glVertex3f(pmin.x-dd/SQRT2,y,pmin.z-dd/SQRT2);
      }
   	glEnd();

   	/* Remaining base plane */
   	glColor3f(border.r,border.g,border.b);
   	glBegin(GL_LINE_STRIP);
   	glVertex3f(pmax.x,0.0,pmin.z);
   	glVertex3f(pmax.x,0.0,pmax.z);
   	glVertex3f(pmin.x,0.0,pmax.z);
   	glEnd();
	}

	if (options.showaxesgrid) {
		glColor3f(0.5,0.5,0.5);

		/* x-z plane */
		glBegin(GL_LINES);
		for (i=0;i<=n;i++) {
			x = pmin.x + i * (pmax.x -pmin.x) / n;
         glVertex3f(x,0.0,pmin.z);
         glVertex3f(x,0.0,pmax.z);  
		}
      for (i=0;i<=n;i++) {
         z = pmin.z + i * (pmax.z -pmin.z) / n;
         glVertex3f(pmin.x,0.0,z);
         glVertex3f(pmax.x,0.0,z);
      }
		glEnd();

      /* x-y plane */
      glBegin(GL_LINES);
      for (i=0;i<=n;i++) {
         x = pmin.x + i * (pmax.x -pmin.x) / n;
         glVertex3f(x,0.0,pmin.z);
         glVertex3f(x,options.scalefactors.y,pmin.z);
      }
      for (i=0;i<=n;i++) {
         y = i * options.scalefactors.y / n;
         glVertex3f(pmin.x,y,pmin.z);
         glVertex3f(pmax.x,y,pmin.z);
      }
      glEnd();

      /* z-y plane */
      glBegin(GL_LINES);
      for (i=0;i<=n;i++) {
         z = pmin.z + i * (pmax.z -pmin.z) / n;
         glVertex3f(pmin.x,0.0,z);
         glVertex3f(pmin.x,options.scalefactors.y,z);
      }
      for (i=0;i<=n;i++) {
         y = i * options.scalefactors.y / n;
         glVertex3f(pmin.x,y,pmin.z);
         glVertex3f(pmin.x,y,pmax.z);
      }
      glEnd();

	}
}

/*
	Optionally draw the bounding box
*/
void DrawBoundingBox(void) 
{
	XYZ p0,p1;

	p0 = pmin;
	p0.y = 0;
	p1 = pmax;
	p1.y = options.scalefactors.y;

   glDisable(GL_LIGHTING);
   glColor3f(1.0,0.0,0.0);
   CreateBox(p0,p1,0);
}

/*
   Set up the lighting environment
*/
void MakeLighting(void)
{
	int i;
	GLfloat globalambient[4] = {0.3,0.3,0.3,1.0};
   GLfloat white[4] = {1.0,1.0,1.0,1.0};
   GLfloat black[4] = {0.0,0.0,0.0,1.0};
	GLfloat grey[4]  = {0.5,0.5,0.5,1.0};
   int deflightlist[8] = {GL_LIGHT0,GL_LIGHT1,GL_LIGHT2,GL_LIGHT3,
                          GL_LIGHT4,GL_LIGHT5,GL_LIGHT6,GL_LIGHT7};
	GLfloat p[4];
	XYZ q;

   /*
      Turn off all the lights
      Leave the ambient component up to the global setting
      Set the specular component appropriately
   */
   for (i=0;i<8;i++) {
      glDisable(deflightlist[i]);
      glLightfv(deflightlist[i],GL_AMBIENT,black);
		glLightfv(deflightlist[i],GL_DIFFUSE,white);
		glLightfv(deflightlist[i],GL_SPECULAR,black);
   }

   /* Rendering mode specific settings */
   switch (options.rendermode) {
   case WIREFRAME:
      glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT,white);
      glDisable(GL_LIGHTING);
      glShadeModel(GL_SMOOTH);
      break;
	case FLAT:
      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT,white);
      glDisable(GL_LIGHTING);
      glShadeModel(GL_FLAT);
		break;
	case SMOOTH:
		glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT,white);
      glDisable(GL_LIGHTING);
      glShadeModel(GL_SMOOTH);
		break;
   case LIGHT:
      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT,globalambient);
      glEnable(GL_LIGHTING);
      glShadeModel(GL_SMOOTH);
      break;
   }

	// Turn on the lights, see drawlights() 
   thelights[0].x = 2*pmax.x;
   thelights[0].y = pmax.y;
   thelights[0].z = 2*pmax.z;
   thelights[1].x = pmin.x;
   thelights[1].y = 2*pmax.y;
   thelights[1].z = pmin.z;
	if (options.rendermode == LIGHT) {

		q = SetPoint(thelights[0]);
		p[0] = q.x;
		p[1] = q.y;
		p[2] = q.z;
		p[3] = 1;
		glLightfv(deflightlist[i],GL_DIFFUSE,white);
   	glLightfv(GL_LIGHT0,GL_POSITION,p);
   	glEnable(GL_LIGHT0);

      q = SetPoint(thelights[1]);
      p[0] = q.x;
      p[1] = q.y;
      p[2] = q.z;
      p[3] = 1;
		glLightfv(deflightlist[i],GL_DIFFUSE,grey);
      glLightfv(GL_LIGHT1,GL_POSITION,p);
      glEnable(GL_LIGHT1);
	}
}

/*
   Deal with plain key strokes
	The x and y are the screen coordinate of the mouse at the time
*/
void HandleKeyboard(unsigned char key,int x, int y)
{
   switch (key) {
   case 27: 									/* ESC 			*/
   case 'Q':									/* Quit 			*/
   case 'q': 
		exit(0); 
		break;
	case 'h':									/* Go home		*/
	case 'H':
		options.autopilot = FALSE;
		options.autospin = FALSE;
		CameraHome(HOME);
		break;
	case '[':                           /* Roll anti clockwise */
		RotateCamera(0.0,0.0,-1.0,1.0);
		break;
	case ']':                           /* Roll clockwise */
		RotateCamera(0.0,0.0,1.0,1.0);
		break;
	case 'i':									/* Translate camera up */
		TranslateCamera(0,1);
		break;
   case 'k':                           /* Translate camera down */
      TranslateCamera(0,-1);
      break;
   case 'j':                           /* Translate camera left */
      TranslateCamera(-1,0);
      break;
   case 'l':                           /* Translate camera right */
      TranslateCamera(1,0);
      break;
	case '-':									/* Move back 	*/
	case '_':
		camera.vp.x -= options.dtheta * camera.focallength * camera.vd.x / 90;
		camera.vp.y -= options.dtheta * camera.focallength * camera.vd.y / 90;
		camera.vp.z -= options.dtheta * camera.focallength * camera.vd.z / 90;
		break;
   case '+':									/* Move forward */
	case '=':
      camera.vp.x += options.dtheta * camera.focallength * camera.vd.x / 90;
      camera.vp.y += options.dtheta * camera.focallength * camera.vd.y / 90;
      camera.vp.z += options.dtheta * camera.focallength * camera.vd.z / 90;
      break;
   case '1':                           /* Special camera views */
   case '2':
   case '3':
   case '4':
   case '5':
   case '6':
   case '7':
      CameraHome((key-'1')+3);
      break;
	case 'w':									/* Write the image to disk */
		options.windowdump = TRUE;
		break;
	case 'W':
		options.dobigdump = TRUE;
		break;
	case 'z':									/* Z axis scale */
	case 'Z':
	   UpdateForm();
      fl_show_form(scaleform->scaleform,FL_PLACE_CENTER,TRUE,NULL);
      scaleformvisible = TRUE;
		break;
   case 'c':                           /* Contour levels */
   case 'C':
      UpdateForm();
      fl_show_form(contourform->contourform,FL_PLACE_CENTER,TRUE,NULL);
      contourformvisible = TRUE;
      break;
	case 'D':									/* Delete all the geometry */
		DeleteData();
		MakeGeometry();
		break;
   }
}

/*
   Deal with special key strokes
	- Arrow keys rotate about the focal position
*/
void HandleSpecialKeyboard(int key,int x, int y)
{
   switch (key) {
   case GLUT_KEY_LEFT:  
		RotateCamera(-1.0,0.0,0.0,1.0); 
		break;
   case GLUT_KEY_RIGHT: 
		RotateCamera(1.0,0.0,0.0,1.0);  
		break;
   case GLUT_KEY_UP:    
		RotateCamera(0.0,1.0,0.0,1.0);  
		break;
   case GLUT_KEY_DOWN:  
		RotateCamera(0.0,-1.0,0.0,1.0); 
		break;
   case GLUT_KEY_F1:
		options.showhelpinfo = !options.showhelpinfo; 
      break;
   case GLUT_KEY_F2:
		options.combineeyes = !options.combineeyes;
      break;
   case GLUT_KEY_F3:
		options.recordimages = !options.recordimages;
      break;
   case GLUT_KEY_F4:
      options.showconstruction = !options.showconstruction;
      break;
   case GLUT_KEY_F5:
		options.showboundingbox = !options.showboundingbox;
      break;
   case GLUT_KEY_F6:
      options.showaxes = !options.showaxes;
      break;
   case GLUT_KEY_F7:
		options.showaxesgrid = !options.showaxesgrid;
      break;
   case GLUT_KEY_F8:
      break;
   case GLUT_KEY_F9:
      break;
   case GLUT_KEY_F10:
      break;
   case GLUT_KEY_F11:
      break;
   case GLUT_KEY_F12:
      break;
   case GLUT_KEY_PAGE_UP:
      break;
   case GLUT_KEY_PAGE_DOWN:
      break;
   case GLUT_KEY_HOME:
      CameraHome(HOME);
      break;
   case GLUT_KEY_END:
      exit(0);
      break;
   case GLUT_KEY_INSERT:
      break;
   }
}

/*
	Rotate (ix,iy) or roll (iz) the camera about the focal point
	ix,iy,iz are flags, 0 do nothing, +- 1 rotates in opposite directions
	Correctly updating all camera attributes
*/
void RotateCamera(double ix,double iy,double iz,double dtheta)
{
   XYZ vp,vu,vd;
   XYZ right;
   XYZ newvp,newr;
   double radius,dd;
   double dx,dy,dz;

   vu = camera.vu;
   Normalise(&vu);
   vp = camera.vp;
   vd = camera.vd;
   Normalise(&vd);
   CROSSPROD(vd,vu,right);
   Normalise(&right);

   /* Handle the roll */
   if (ABS(iz) > 0.0001) {
      camera.vu.x += iz * right.x * dtheta * 5 * DTOR;
      camera.vu.y += iz * right.y * dtheta * 5 * DTOR;
      camera.vu.z += iz * right.z * dtheta * 5 * DTOR;
      Normalise(&camera.vu);
      return;
   }

   /* Distance from the rotate point */
   dx = camera.vp.x - camera.pr.x;
   dy = camera.vp.y - camera.pr.y;
   dz = camera.vp.z - camera.pr.z;
   radius = sqrt(dx*dx + dy*dy + dz*dz);

   /* Determine the new view point */
   dd = radius * dtheta * DTOR * 5;
   newvp.x = vp.x + dd * ix * right.x + dd * iy * vu.x - camera.pr.x;
   newvp.y = vp.y + dd * ix * right.y + dd * iy * vu.y - camera.pr.y;
   newvp.z = vp.z + dd * ix * right.z + dd * iy * vu.z - camera.pr.z;
   Normalise(&newvp);
   camera.vp.x = camera.pr.x + radius * newvp.x;
   camera.vp.y = camera.pr.y + radius * newvp.y;
   camera.vp.z = camera.pr.z + radius * newvp.z;

   /* Determine the new right vector */
   newr.x = camera.vp.x + right.x - camera.pr.x;
   newr.y = camera.vp.y + right.y - camera.pr.y;
   newr.z = camera.vp.z + right.z - camera.pr.z;
   Normalise(&newr);
   newr.x = camera.pr.x + radius * newr.x - camera.vp.x;
   newr.y = camera.pr.y + radius * newr.y - camera.vp.y;
   newr.z = camera.pr.z + radius * newr.z - camera.vp.z;
   Normalise(&newr);

   camera.vd.x = camera.pr.x - camera.vp.x;
   camera.vd.y = camera.pr.y - camera.vp.y;
   camera.vd.z = camera.pr.z - camera.vp.z;
   Normalise(&camera.vd);

   /* Determine the new up vector */
   CROSSPROD(newr,camera.vd,camera.vu);
   Normalise(&camera.vu);
}

/*
   Translate the camera view point
*/
void TranslateCamera(int ix,int iy)
{
   XYZ vp,vu,vd;
   XYZ right;
   double radians;

   vu = camera.vu;
   Normalise(&vu);
   vp = camera.vp;
   vd = camera.vd;
   Normalise(&vd);
   CROSSPROD(vd,vu,right);
   Normalise(&right);
   radians = options.dtheta * PI / 180.0;
	
	camera.vp.x += iy * vu.x * options.dtheta * camera.focallength / 90;
   camera.vp.y += iy * vu.y * options.dtheta * camera.focallength / 90;
	camera.vp.z += iy * vu.z * options.dtheta * camera.focallength / 90; 
   camera.pr.x += iy * vu.x * options.dtheta * camera.focallength / 90;
   camera.pr.y += iy * vu.y * options.dtheta * camera.focallength / 90;
   camera.pr.z += iy * vu.z * options.dtheta * camera.focallength / 90;

   camera.vp.x += ix * right.x * options.dtheta * camera.focallength / 90;
   camera.vp.y += ix * right.y * options.dtheta * camera.focallength / 90;
   camera.vp.z += ix * right.z * options.dtheta * camera.focallength / 90;
   camera.pr.x += ix * right.x * options.dtheta * camera.focallength / 90;
   camera.pr.y += ix * right.y * options.dtheta * camera.focallength / 90;
   camera.pr.z += ix * right.z * options.dtheta * camera.focallength / 90;
}

/*
   Handle mouse events
	Right button events are passed to the menu handler
	Reduce the resolution on mouse interaction
*/
void HandleMouse(int button,int state,int x,int y)
{
   if (state == GLUT_DOWN) {
      if (button == GLUT_LEFT_BUTTON) {
			options.currentbutton = GLUT_LEFT_BUTTON;			
      } else if (button == GLUT_MIDDLE_BUTTON) {
			options.currentbutton = GLUT_MIDDLE_BUTTON;
      }
   } else if (state == GLUT_UP) {
      
	}
}

/*
   Handle mouse motion
*/
void HandleMouseMotion(int x,int y)
{
   static int xlast=-1,ylast=-1;
	double dx,dy;

	dx = x - xlast;
	dy = y - ylast;
	if (dx < 0)      dx = -1;
	else if (dx > 0) dx =  1;
	if (dy < 0)      dy = -1;
	else if (dy > 0) dy =  1;

	if (options.currentbutton == GLUT_LEFT_BUTTON)
  		RotateCamera(-dx,dy,0.0,0.5);
	else if (options.currentbutton == GLUT_MIDDLE_BUTTON)
		RotateCamera(0.0,0.0,dx,0.5);

   xlast = x;
   ylast = y;
}

/*
	Move the 3D mouse around on passive mouse movement events
*/
void HandlePassiveMotion(int x,int y)
{
	static int xlast=-1,ylast=-1;
	int dx,dy;
   XYZ vp,vu,vd;
   XYZ right;

   vu = camera.vu;
   Normalise(&vu);
   vp = camera.vp;
   vd = camera.vd;
   Normalise(&vd);
   CROSSPROD(vd,vu,right);
   Normalise(&right);

   dx = x - xlast;
   dy = y - ylast;
   if (dx < 0)      dx = -1;
   else if (dx > 0) dx =  1;
   if (dy < 0)      dy = -1;
   else if (dy > 0) dy =  1;

	xlast = x;
	ylast = y;
}

/*
   How to handle visibility
*/
void HandleVisibility(int visible)
{
   if (visible == GLUT_VISIBLE)
      glutIdleFunc(HandleIdle);
   else
      glutIdleFunc(NULL);
}

/*
   What to do on idle events
*/
void HandleIdle(void)
{
   static double tstart=-1,tstop;
	FL_OBJECT *retobj = NULL;

	if (stereoformvisible || scaleformvisible || contourformvisible) {
		retobj = fl_check_forms();

		if (stereoformvisible) {
   		if (retobj == stereoform->Hide) {
   	   	fl_hide_form(stereoform->stereoform);
				stereoformvisible = FALSE;
			}
			if (retobj == stereoform->Eyesep) {
				camera.eyesep = fl_get_slider_value(stereoform->Eyesep);
			}
	   	if (retobj == stereoform->Focallength) {
	      	camera.focallength = fl_get_slider_value(stereoform->Focallength);
	   	}
			if (retobj == stereoform->Aperture) {
				camera.aperture = fl_get_slider_value(stereoform->Aperture);
			}
		}
		if (scaleformvisible) {
			if (retobj == scaleform->Hide) {
				fl_hide_form(scaleform->scaleform);
				scaleformvisible = FALSE;
			}
			if (retobj == scaleform->ZScale) {
				options.scalefactors.y = atof(fl_get_input(scaleform->ZScale));
				MakeGeometry();
			}
		}
	   if (contourformvisible) {
	      if (retobj == contourform->Hide) {
	         fl_hide_form(contourform->contourform);
	         contourformvisible = FALSE;
	      }
	      if (retobj == contourform->Ncontour) {
	         options.ncontour = atof(fl_get_input(contourform->Ncontour));
	         MakeGeometry();
	      }
		}
   }

	/* Is it time for an update */
	if (tstart < 0) 
		tstart = GetRunTime();
	tstop = GetRunTime();
	if (tstop - tstart > 1.0/options.targetfps) {
      // Nasty hack to solve Mac OS-X cursor problem
      if (options.stereo == DUALSTEREO) {
         glutSetCursor(GLUT_CURSOR_CROSSHAIR);
         glutSetCursor(GLUT_CURSOR_NONE);
      }
   	glutPostRedisplay();
		tstart = GetRunTime();
	}
}

/*
	Set the camera to the "home" position
*/
void CameraHome(int mode)
{
	double size;
	XYZ center;

	ComputeBounds();

	/* The center of the model */
	center = SetPoint(pmid);

	size = MAX((pmax.x-pmin.x),(pmax.z-pmin.z));
	size = MAX(size,options.scalefactors.y);

	/* Set the camera attributes */
	camera.pr = center;
	camera.vp = center;
	switch (mode) {
	case HOME:
		camera.aperture = 45;
		camera.focallength = 1.75*size;
   	camera.eyesep = size / 20.0;
		camera.vp.z += 1.75*size;
		camera.vu.x = 0; camera.vu.y = 1; camera.vu.z = 0;
		break;
   case FRONT:
      camera.vp.z += 1.75 * size;
      camera.vu.x = 0; camera.vu.y = 1; camera.vu.z = 0;
      break;
   case BACK:
      camera.vp.z -= 1.75 * size;
      camera.vu.x = 0; camera.vu.y = 1; camera.vu.z = 0;
      break;
   case LEFT:
      camera.vp.x -= 1.75 * size;
      camera.vu.x = 0; camera.vu.y = 1; camera.vu.z = 0;
      break;
   case RIGHT:
      camera.vp.x += 1.75 * size;
      camera.vu.x = 0; camera.vu.y = 1; camera.vu.z = 0;
      break;
   case TOP:
      camera.vp.y += 1.75 * size;
      camera.vu.x = 0; camera.vu.y = 0; camera.vu.z = 1;
      break;
   case BOTTOM:
      camera.vp.y -= 1.75 * size;
      camera.vu.x = 0; camera.vu.y = 0; camera.vu.z = 1;
      break;
   case AXONOMETRIC:
      camera.vp.x -= size;
      camera.vp.y += size;
      camera.vp.z -= size;
      camera.vu.x = 0; camera.vu.y = 1; camera.vu.z = 0;
      break;
	}

   /* Point the camera toward the mid point of the model */
   camera.vd.x = center.x - camera.vp.x;
   camera.vd.y = center.y - camera.vp.y;
   camera.vd.z = center.z - camera.vp.z;

	/* Update the slider bounds */
	fl_set_slider_bounds(stereoform->Eyesep,0.0,size/5);
	fl_set_slider_bounds(stereoform->Aperture,0.0,90.0);
   fl_set_slider_bounds(stereoform->Focallength,size/2,6*size);

	if (options.debug) {
		fprintf(stderr,"Bounds = (%g,%g,%g) -> (%g,%g,%g)\n",pmin.x,pmin.z,pmin.y,pmax.x,pmax.z,pmax.y);
		fprintf(stderr,"Camera position = (%g,%g,%g)\n",camera.vp.x,camera.vp.z,camera.vp.y);
      fprintf(stderr,"Camera view direction = (%g,%g,%g)\n",camera.vd.x,camera.vd.z,camera.vd.y);
      fprintf(stderr,"Camera aperture = %g\n",camera.aperture);
      fprintf(stderr,"Camera focallength = %g\n",camera.focallength);
      fprintf(stderr,"Camera eyesep = %g\n",camera.eyesep);

	}

	UpdateForm();
}

/*
	Set the auto camera position
	The first time, get the file name and read the flight path
*/
int AutoPilot(int mode)
{
   const char *fn;
   char fname[256];
	static FILE *fptr = NULL;
	XYZ vp,vd,vu;
	
	if (!options.autopilot)
		return(FALSE);

	/* Are we getting a new flight path ? */
	if (mode == 0) {
		if (fptr != NULL)
			fclose(fptr);

   	fn = fl_show_file_selector("Flight path file","","","");
   	if (fn == NULL || strlen(fn) <= 0) {
      	fname[0] = '\0';
			fptr = NULL;
      	return(FALSE);
   	}
   	strcpy(fname,fn);
		if ((fptr = fopen(fname,"r")) == NULL) {
			fptr = NULL;
			return(FALSE);
		}
	}

	/* Get the next camera position */
	if (fscanf(fptr,"%lf %lf %lf %lf %lf %lf %lf %lf %lf",
		&vp.x,&vp.z,&vp.y, &vd.x,&vd.z,&vd.y, &vu.x,&vu.z,&vu.y) != 9) {
		fclose(fptr);
		fptr = NULL;
		return(FALSE);
	}
	camera.vp = vp; Normalise(&vd);
	camera.vd = vd; Normalise(&vd);
	camera.vu = vu; Normalise(&vu);

	return(TRUE);
}

/*
	Update the slider on the stereo settings form
*/
void UpdateForm(void)
{
	char s[64];

	fl_set_slider_value(stereoform->Eyesep,camera.eyesep);
	fl_set_slider_value(stereoform->Aperture,camera.aperture);
	fl_set_slider_value(stereoform->Focallength,camera.focallength);

	sprintf(s,"%g",options.scalefactors.y);
	fl_set_input(scaleform->ZScale,s); 

   sprintf(s,"%d",options.ncontour);
   fl_set_input(contourform->Ncontour,s);
}

