#include "glgraph.h"

extern OPTIONS options;
extern CAMERA camera;

extern int nx;
extern int ny;
extern XYZ *grid;
extern int ntri;
extern ITRIANGLE *triangles;

void SaveViewFile(CAMERA c)
{
   const char *fn;
   char fname[256];
   FILE *fptr;

   fn = fl_show_file_selector("View file","","","");
   if (fn == NULL || strlen(fn) <= 0)
      return;
   strcpy(fname,fn);
   if ((fptr = fopen(fname,"w")) == NULL)
      return;
   fprintf(fptr,"vp %g %g %g\n",c.vp.x,c.vp.y,c.vp.z);
   fprintf(fptr,"vd %g %g %g\n",c.vd.x,c.vd.y,c.vd.z);
   fprintf(fptr,"vu %g %g %g\n",c.vu.x,c.vu.y,c.vu.z);
   fprintf(fptr,"pr %g %g %g\n",c.pr.x,c.pr.y,c.pr.z);
   fprintf(fptr,"focallength %g\n",c.focallength);
   fprintf(fptr,"aperture %g\n",c.aperture);
   fprintf(fptr,"eyeseparation %g\n",c.eyesep);
   fclose(fptr);
}

void ReadViewFile(char *name,CAMERA *c)
{
   const char *fn;
   char fname[256];
   char s[64];
   FILE *fptr;

   if (strlen(name) < 1) {
      fn = fl_show_file_selector("View file","","","");
      if (fn == NULL || strlen(fn) <= 0)
         return;
      strcpy(fname,fn);
   } else {
      strcpy(fname,name);
   }
   if ((fptr = fopen(fname,"r")) == NULL)
      return;
   while (fscanf(fptr,"%s",s) == 1) {
      if (strcmp(s,"vp") == 0)
         fscanf(fptr,"%lf %lf %lf",&(c->vp.x),&(c->vp.y),&(c->vp.z));
      if (strcmp(s,"vd") == 0)
         fscanf(fptr,"%lf %lf %lf",&(c->vd.x),&(c->vd.y),&(c->vd.z));
      if (strcmp(s,"vu") == 0)
         fscanf(fptr,"%lf %lf %lf",&(c->vu.x),&(c->vu.y),&(c->vu.z));
      if (strcmp(s,"pr") == 0)
         fscanf(fptr,"%lf %lf %lf",&(c->pr.x),&(c->pr.y),&(c->pr.z));
      if (strcmp(s,"focallength") == 0)
         fscanf(fptr,"%lf",&(c->focallength));
      if (strcmp(s,"aperture") == 0)
         fscanf(fptr,"%lf",&(c->aperture));
      if (strcmp(s,"eyesep") == 0)
         fscanf(fptr,"%lf",&(c->eyesep));
   }
   fclose(fptr);
}

/*
   Read a file of random samples, and optionally triangulate
   The file is simply a list of x,y,z triples
*/
int ReadRandom(char *name,int triang)
{
   int i,j,ndup=0,anerror;
   double z;
   double xmin = 1e32,ymin = 1e32,xmax = -1e32,ymax = -1e32;
   XYZ p;
   FILE *fptr;
   const char *fn;
   char fname[256];
   short *flag = NULL;

   /* Was a file name supplied or do we prompt for one */
   if (strlen(name) < 1) {
      fn = fl_show_file_selector("Random point file name","","","");
      if (fn == NULL || strlen(fn) <= 0)
         return(TRUE);
      strcpy(fname,fn);
   } else {
      strcpy(fname,name);
   }

   /* Open the file */
   if ((fptr = fopen(fname,"r")) == NULL) {
      fl_show_message("File Error","Unable to open sample file","");
      return(FALSE);
   }
   if (options.debug)
      fprintf(stderr,"Opened random point file OK\n");

   /* Read until the end of file is reached */
   nx = 0;
   ny = 1;
   while (fscanf(fptr,"%lf %lf %lf",&p.x,&p.y,&p.z) == 3) {
      if ((grid = realloc(grid,(nx+10)*sizeof(XYZ))) == NULL) {
         fl_show_message("Memory error","Vertex realloc failed","");
         DeleteData();
         fclose(fptr);
         return(FALSE);
      }
      grid[nx++] = p;
      xmin = MIN(xmin,p.x);
      ymin = MIN(ymin,p.y);
      xmax = MAX(xmax,p.x);
      ymax = MAX(ymax,p.y);
   }
   fclose(fptr);
   if (options.debug) {
      fprintf(stderr,"Read %d points\n",nx);
      fprintf(stderr,"X range = %g to %g\n",xmin,xmax);
      fprintf(stderr,"Y range = %g to %g\n",ymin,ymax);
   }

   /* Do we have enough points */
   if (nx < 5) {
      fl_show_message("File error","Not enough points","Need > 4");
      DeleteData();
      return(FALSE);
   }

   if (triang) {

      /* Create space for triangles */
      if ((triangles = malloc(3*nx*sizeof(ITRIANGLE))) == NULL) {
         fl_show_message("Memory error","Triangle malloc failed","");
         DeleteData();
         return(FALSE);
      }

      /* Check for coincident values */
      flag = malloc(nx*sizeof(short));
      qsort(grid,nx,sizeof(XYZ),XYZCompare);
      if (options.debug)
         fprintf(stderr,"Checking for coincident values\n");
      for (i=0;i<nx;i++)
         flag[i] = 0;
      ndup = 0;
      for (i=0;i<nx-1;i++) {
         if (flag[i] != 0)
            continue;
         for (j=i+1;j<nx;j++) {
            if (flag[j] != 0)
               continue;
            if (ABS(grid[i].x-grid[j].x) < EPSILON) {
               if (ABS(grid[i].y-grid[j].y) < EPSILON) {
                  ndup++;
                  flag[j] = 1;
               }
            } else {
               break;
            }
         }
      }

      /* Remove the coincident points */
      if (options.debug)
         fprintf(stderr,"Removing coincident points\n");
      for (i=0;i<nx;i++) {
         if (flag[i] != 0) {
            grid[i] = grid[nx-1];
            flag[i] = flag[nx-1];
            nx--;
            i--;
         }
      }
      free(flag);
      if (options.debug)
         fprintf(stderr,"Found %d duplicate points\n",ndup);

      /* Sort the samples in x */
      if (options.debug)
         fprintf(stderr,"Sorting\n");
      qsort(grid,nx,sizeof(XYZ),XYZCompare);

      /* Do the triangulation */
      if (options.debug)
         fprintf(stderr,"Triangulating\n");
      if ((anerror = Triangulate(nx,grid,triangles,&ntri)) > 0) {
         fprintf(stderr,"Triangulate failed, fail code %d\n",anerror);
         DeleteData();
         return(FALSE);
      }
      if (options.debug)
         fprintf(stderr,"Created %d triangles\n",ntri);
   }

   /* Flip the y and z coordinates for local convention (y is up instead of z) */
   for (i=0;i<nx;i++) {
      z = grid[i].y;
      grid[i].y = grid[i].z;
      grid[i].z = z;
   }

   return(TRUE);
}

/*
   Read an grid file
   The file is structures as follows
   First line:  nx ny
   Second line: x axis divisions
   Third line:  y axis divisions
   Remaining:   grid points, nx*ny of them
	Second and third bits are missing for second grid type
*/
int ReadGrid(char *name)
{
   int i,j;
   double x,y,z;
   const char *fn;
   char fname[256];
   FILE *fptr;

   /* Do we need to prompt for a file name */
   if (strlen(name) < 1) {
      fn = fl_show_file_selector("Grid file name","","","");
      if (fn == NULL || strlen(fn) <= 0)
         return(TRUE);
      strcpy(fname,fn);
   } else {
      strcpy(fname,name);
   }
   if (options.debug)
      fprintf(stderr,"Attempting to read file \"%s\"\n",fname);

	// Reset various gllobals
	options.scalefactors.x = 1; 
	options.scalefactors.y = 1; 
	options.scalefactors.z = 1;

   if ((fptr = fopen(fname,"r")) == NULL) {
      fl_show_message("File Error","Unable to open array file","");
      return(FALSE);
   }
   if (fscanf(fptr,"%d %d",&nx,&ny) != 2 || nx < 2 || ny < 2) {
      fl_show_message("File Error","Failed to read legal grid size","");
      DeleteData();
      fclose(fptr);
      return(FALSE);
   }
	if (options.debug)
		fprintf(stderr,"Grid size: %d x %d\n",nx,ny);

   if ((grid = malloc(nx*ny*sizeof(XYZ))) == NULL) {
      DeleteData();
      fl_show_message("File Error","Failed to allocate memory for grid","");
      fclose(fptr);
      return(FALSE);
   }

	// Zero heights
	for (i=0;i<nx;i++) 
      for (j=0;j<ny;j++)
			grid[j*nx+i].y = 0;

	// Set or read cell positions
	if (options.datatype == GRIDDED2) {
		for (i=0;i<nx;i++) {
			for (j=0;j<ny;j++) {
				grid[j*nx+i].x = i;
				grid[j*nx+i].z = j;
			}
		}
	} else {
   	for (i=0;i<nx;i++) {
   	   if (fscanf(fptr,"%lf",&x) != 1) {
   	      DeleteData();
   	      fl_show_message("File Error","Failed to read x axis","");
   	      fclose(fptr);
   	      return(FALSE);
   	   }
   	   for (j=0;j<ny;j++)
   	      grid[j*nx+i].x = x;
   	}
   	for (j=0;j<ny;j++) {
   	   if (fscanf(fptr,"%lf",&y) != 1) {
   	      DeleteData();
   	      fl_show_message("File Error","Failed to read y axis","");
   	      fclose(fptr);
   	      return(FALSE);
   	   }
   	   for (i=0;i<nx;i++)
   	      grid[j*nx+i].z = y;
   	}
	}
	/*
	if (options.debug) {
		fprintf(stderr,"X grid points: ");
		for (i=0;i<nx;i++)
			fprintf(stderr,"%g ",grid[i].x);
		fprintf(stderr,"\n");
      fprintf(stderr,"Y grid points: ");
      for (j=0;j<ny;j++)
         fprintf(stderr,"%g ",grid[j*nx].z);
      fprintf(stderr,"\n");
	}
	*/

   // Read the grid 
	for (j=0;j<ny;j++) {
   	for (i=0;i<nx;i++) {
      	if (fscanf(fptr,"%lf",&z) != 1) {
      	   DeleteData();
      	   fl_show_message("File Error","Failed to read height values","");
      	   fclose(fptr);
      	   return(FALSE);
      	}
      	grid[j*nx+i].y = z;
		}
   }
   fclose(fptr);

	/*
   if (options.debug) {
      fprintf(stderr,"Z grid points:\n");
		for (j=0;j<ny;j++) {
      	for (i=0;i<nx;i++) {
         	fprintf(stderr,"(%g,%g,%g) ",grid[j*nx+i].x,grid[j*nx+i].z,grid[j*nx+i].y);
			}
      	fprintf(stderr,"\n");
		}
	}
	*/

	ComputeBounds();

   return(TRUE);
}

