#include "glgraph.h"
#include "usagestring.h"

extern OPTIONS options;
extern CAMERA camera;

extern int nx;
extern int ny;
extern XYZ *grid;
extern int ntri;
extern ITRIANGLE *triangles;
extern XYZ pmin,pmax,pmid;
extern XYZ thelights[2];

/*
   Choose a colour based up glogals and height
*/
COLOUR SetColour(XYZ p)
{
   COLOUR colour;
	static COLOUR last;

   if (options.colourmap == 100)
      colour = options.constcolour;
   else
      colour = GetColour(p.y,pmin.y,pmax.y,options.colourmap);
   if (colour.r != last.r || colour.g != last.g || colour.b != last.b) {
      glColor4f(colour.r,colour.g,colour.b,1.0);
      last = colour;
   }

   return(colour);
}

/*
   Transform a point into prefered coordinates
   Currently just z axis scaling
   After this the dataset ranges in height from 0 to scalefactors.y
*/
XYZ SetPoint(XYZ p)
{
   XYZ newp;

   newp.x = p.x;
   newp.y = (p.y - pmin.y) / (pmax.y - pmin.y) * options.scalefactors.y;
   newp.z = p.z;

   return(newp);
}

/*
   Set up the material attributes
*/
void MakeMaterial(void)
{
   GLfloat spec[4] = {1.0,1.0,1.0,1.0};
   GLfloat shin[1] = {0.0};
   GLfloat emis[4] = {0.0,0.0,0.0,1.0};

   switch (options.rendermode) {
	case WIREFRAME:
	case FLAT:
   case SMOOTH:
	case LIGHT:
      glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,spec);
      glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shin);
      glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,emis);
      break;
   }
}

/*
   Optionally draw the light sources
*/
void DrawLights(void)
{
   double r;
	XYZ p1,p2;

   r = MAX(pmax.x-pmin.x,pmax.z-pmin.z) / 100;
   glColor3f(1.0,1.0,1.0);
	p1 = SetPoint(thelights[0]);
   CreateSphere(p1,r,12,0,0.0,TWOPI,-PID2,PID2);
   glColor3f(0.5,0.5,0.5);
	p2 = SetPoint(thelights[1]);
   CreateSphere(p2,r,12,0,0.0,TWOPI,-PID2,PID2);
}

/*
   Calculate the normal of a grid vertex
*/
XYZ GridNormal(int i,int j)
{
   int index1,index2,index3;
   XYZ p1,p2,p3,normal;

   index1 = j * nx + i;
   index2 = j * nx + i + 1;
   index3 = (j+1) * nx + i;
   if (i == nx-1) {
      index2 = (j+1) * nx + i;
      index3 = j * nx + i - 1;
   }
   if (j == ny-1) {
      index2 = (j-1) * nx + i;
      index3 = j * nx + i + 1;
   }
   if (i == nx-1 && j == ny-1) {
      index2 = j * nx + i - 1;
      index3 = (j-1) * nx + i;
   }

   p1 = SetPoint(grid[index1]);
   p2 = SetPoint(grid[index2]);
   p3 = SetPoint(grid[index3]);
   normal = CalcNormal(p1,p2,p3);

   return(normal);
}

/*
   Display the program usage information
*/
void GiveUsage(char *cmd)
{
   fprintf(stderr,"Usage: %s [command line options]\n",cmd);
   fprintf(stderr,"%s\n",usagestring);
   exit(-1);
}

/* 
   Delete/free all the data structures
*/
void DeleteData(void)
{  
   if (grid != NULL)       { free(grid );     grid  = NULL;     }
   if (triangles != NULL)  { free(triangles); triangles = NULL; }
   nx = 0;
   ny = 0;
   ntri = 0;
   options.datatype = NODATA;
   if (options.debug)
      fprintf(stderr,"Data deleted\n");
}

/* 
   Update model bounds
*/
void ComputeBounds(void)
{  
   int i;
   
   /* Find the bounds of the geometry */
   pmin.x = 1e32;  pmin.y = 1e32;  pmin.z = 1e32;
   pmax.x = -1e32; pmax.y = -1e32; pmax.z = -1e32;
   
   if (options.datatype == GRIDDED || options.datatype == GRIDDED2 || options.datatype == POINTED) {
      for (i=0;i<nx*ny;i++) {
         MinMaxXYZ(grid[i],&pmin,&pmax);
			//fprintf(stderr,"(%g %g %g) ",grid[i].x,grid[i].z,grid[i].y);
		}
   } else if (options.datatype == TRIANGULAR) {
      for (i=0;i<ntri;i++) {
         MinMaxXYZ(grid[triangles[i].p1],&pmin,&pmax);
         MinMaxXYZ(grid[triangles[i].p2],&pmin,&pmax);
         MinMaxXYZ(grid[triangles[i].p3],&pmin,&pmax);
      }
   } else {
      pmin.x = -1; pmin.y = -1; pmin.z = -1;
      pmax.x =  1; pmax.y =  1; pmax.z =  1;
   }
   
   if (pmax.x <= pmin.x) { pmax.x += 0.01; pmin.x -= 0.01; }
   if (pmax.y <= pmin.y) { pmax.y += 0.01; pmin.y -= 0.01; }
   if (pmax.z <= pmin.z) { pmax.z += 0.01; pmin.z -= 0.01; }

   pmid.x = (pmin.x + pmax.x) / 2;
   pmid.y = (pmin.y + pmax.y) / 2;
   pmid.z = (pmin.z + pmax.z) / 2;

	if (options.debug)
		fprintf(stderr,"Bounds = (%g,%g,%g) -> (%g,%g,%g)\n",pmin.x,pmin.z,pmin.y,pmax.x,pmax.z,pmax.y);
}

int XYZCompare(const void *v1,const void *v2)
{
   XYZ *p1,*p2;

   p1 = (XYZ *)v1;
   p2 = (XYZ *)v2;

   if (p1->x < p2->x)
      return(-1);
   else if (p1->x > p2->x)
      return(1);
   else
      return(0);
}

/*
	Draw a box for the boxed plats
*/
void DrawABox(XYZ pmin,XYZ pmax)
{
   glBegin(GL_QUADS);
   glNormal3f(0.0,0.0,-1.0);
   glVertex3f(pmin.x,pmin.y,pmin.z);
   glVertex3f(pmin.x,pmax.y,pmin.z);
   glVertex3f(pmax.x,pmax.y,pmin.z);
   glVertex3f(pmax.x,pmin.y,pmin.z);
   glNormal3f(0.0,0.0,1.0);
   glVertex3f(pmin.x,pmin.y,pmax.z);
   glVertex3f(pmax.x,pmin.y,pmax.z);
   glVertex3f(pmax.x,pmax.y,pmax.z);
   glVertex3f(pmin.x,pmax.y,pmax.z);
   glNormal3f(-1.0,0.0,0.0);
   glVertex3f(pmin.x,pmin.y,pmax.z);
   glVertex3f(pmin.x,pmax.y,pmax.z);
   glVertex3f(pmin.x,pmax.y,pmin.z);
   glVertex3f(pmin.x,pmin.y,pmin.z);
   glNormal3f(1.0,0.0,0.0);      
   glVertex3f(pmax.x,pmin.y,pmax.z);
   glVertex3f(pmax.x,pmin.y,pmin.z);
   glVertex3f(pmax.x,pmax.y,pmin.z);
   glVertex3f(pmax.x,pmax.y,pmax.z);
   glNormal3f(0.0,-1.0,0.0);
   glVertex3f(pmin.x,pmin.y,pmax.z);
   glVertex3f(pmin.x,pmin.y,pmin.z);
   glVertex3f(pmax.x,pmin.y,pmin.z);
   glVertex3f(pmax.x,pmin.y,pmax.z);
   glNormal3f(0.0,1.0,0.0);
   glVertex3f(pmin.x,pmax.y,pmax.z);
   glVertex3f(pmax.x,pmax.y,pmax.z);
   glVertex3f(pmax.x,pmax.y,pmin.z);
   glVertex3f(pmin.x,pmax.y,pmin.z);
   glEnd();
}

